@@ -195,7 +195,7 @@ impl std::fmt::Display for Capability {
             Capability::CAP_CHECKPOINT_RESTORE => "cap_checkpoint_restore",
             Capability::__Nonexhaustive => unreachable!("invalid capability"),
         };
-        write!(f, "{}", name)
+        write!(f, "{name}")
     }
 }
 
