@@ -32,7 +32,7 @@ use btoi::btoi;
 use nix::{
     errno::Errno,
     fcntl::OFlag,
-    sys::stat::Mode,
+    sys::{personality::Persona, stat::Mode},
     unistd::{close, sysconf, SysconfVar},
 };
 // TODO: LazyCell is rust>=1.80.0
@@ -41,7 +41,7 @@ use once_cell::sync::Lazy;
 use crate::{
     check_cross_memory_attach,
     compat::{STATX_MNT_ID, STATX_MNT_ID_UNIQUE, STATX_MODE},
-    fs::{duprand, FileMap, FileMapEntry, FileType},
+    fs::{duprand, is_dev_null, FileMap, FileMapEntry, FileType},
     fstatx, info,
     path::{XPath, XPathBuf},
     proc::proc_mmap_min_addr,
@@ -106,9 +106,6 @@ pub const ENV_PROXY_PORT: &str = "SYD_PROXY_PORT";
 /// The environment variable to read the default external unix socket from.
 pub const ENV_PROXY_UNIX: &str = "SYD_PROXY_UNIX";
 
-/// The environment variable to set to disable logging to syslog.
-pub const ENV_NO_SYSLOG: &str = "SYD_NO_SYSLOG";
-
 /// The environment variable to set to disable cross memory attach, and fallback to /proc/pid/mem.
 pub const ENV_NO_CROSS_MEMORY_ATTACH: &str = "SYD_NO_CROSS_MEMORY_ATTACH";
 
@@ -124,9 +121,6 @@ pub const ENV_DUMP_SCMP: &str = "SYD_DUMP_SCMP";
 /// The environment variable to skip seccomp confinement of Syd threads.
 pub const ENV_SKIP_SCMP: &str = "SYD_SKIP_SCMP";
 
-/// The environment variable to enable seccomp synchronous mode.
-pub const ENV_SYNC_SCMP: &str = "SYD_SYNC_SCMP";
-
 /// The environment variable to set for quick boot.
 ///
 /// This makes Syd startup noticably faster,
@@ -260,7 +254,7 @@ pub static EMU_MAX_SIZE: Lazy<usize> = Lazy::new(|| {
 /// outside the core pool remain alive while not receiving any work
 /// before giving up and terminating.
 /// Defaults to 3 seconds.
-pub const EMU_KEEP_ALIVE: u16 = 3000;
+pub const EMU_KEEP_ALIVE: u16 = 3;
 
 /// Specify the graceful wait time for the monitor thread.
 /// Defaults to 250 milliseconds.
@@ -283,8 +277,8 @@ pub const AES_CYCLE_TIME: Duration = Duration::from_millis(10);
 pub const EMU_STACK_SIZE: usize = 208 * 1024;
 
 /// Stack size for the syscall monitor thread.
-/// Defaults to 96k.
-pub const MON_STACK_SIZE: usize = 96 * 1024;
+/// Defaults to 128k.
+pub const MON_STACK_SIZE: usize = 128 * 1024;
 
 /// Stack size for the syscall interrupt thread.
 /// Defaults to 128k.
@@ -367,10 +361,10 @@ pub const SEGVGUARD_MAXCRASHES: u8 = 5;
 
 // Sandbox Restrictions
 
-/// List of allowlisted personality requests.
+/// List of allowlisted personality(2) flags.
 ///
-/// This is consistent with podman and docker.
-pub const ALLOWLIST_PERSONALITY: &[KeyValue] = &[
+/// This is consistent with podman(1) and docker(1)
+pub const SAFE_PERSONAS: &[KeyValue] = &[
     ("PER_LINUX", 0x0000),
     ("PER_LINUX32", 0x0800000),
     ("UNAME26", 0x0020000),
@@ -378,6 +372,9 @@ pub const ALLOWLIST_PERSONALITY: &[KeyValue] = &[
     ("GET_PERSONALITY", 0xffffffff),
 ];
 
+/// List of unsafe personality(2) flags.
+pub const UNSAFE_PERSONAS: &[Persona] = &[Persona::READ_IMPLIES_EXEC, Persona::ADDR_NO_RANDOMIZE];
+
 /// List of allowlisted prctl requests.
 pub const ALLOWLIST_PRCTL: &[KeyValue] = &[
     ("PR_SET_PDEATHSIG", 1),
@@ -445,7 +442,8 @@ pub const ALLOWLIST_PRCTL: &[KeyValue] = &[
     ("PR_SET_VMA", 0x53564d41),
 ];
 
-/// Equivalent to the list of environment variables that gl*bc removes for SUID programs.
+/// Equivalent to the list of environment variables that glibc/musl remove
+/// for programs running under secure-execution mode.
 pub const UNSAFE_ENV: &[&[u8]] = &[
     b"GCONV_PATH",
     b"GETCONF_DIR",
@@ -478,6 +476,7 @@ pub const UNSAFE_ENV: &[&[u8]] = &[
     b"LOCALDOMAIN",
     b"LOCPATH",
     b"MALLOC_TRACE",
+    b"MUSL_LOCPATH",
     b"NIS_PATH",
     b"NLSPATH",
     b"RESOLV_HOST_CONF",
@@ -536,32 +535,37 @@ pub const UNSAFE_ENV: &[&[u8]] = &[
  * 12. FS_IOC32_GETFLAGS, FS_IOC32_SETFLAGS are forwarded to their device
  *     implementations.
  */
-#[allow(clippy::unnecessary_cast)]
-pub const DEFAULT_IOCTL: &[(u64, bool)] = &[
-    (libc::FIOCLEX as u64, false),
-    (libc::FIONCLEX as u64, false),
-    (libc::FIONBIO as u64, false),
-    (libc::FIOASYNC as u64, false),
-    (crate::fs::FIOQSIZE as u64, false),
-    (crate::fs::FIFREEZE as u64, true),      // CAP_SYS_ADMIN!
-    (crate::fs::FITHAW as u64, true),        // ditto!
-    (crate::fs::FS_IOC_FIEMAP as u64, true), // Leaks on-disk layout.
-    (crate::fs::FIGETBSZ as u64, true),      // ditto!
-    (libc::FICLONE as u64, false),
-    (libc::FICLONERANGE as u64, false),
-    (crate::fs::FIDEDUPERANGE as u64, false),
-    (crate::fs::FS_IOC_GETFSUUID as u64, true), // Leaks on-disk layout.
-    (crate::fs::FS_IOC_GETFSSYSFSPATH as u64, true), // ditto!
-    (crate::fs::FIBMAP as u64, true),           // ditto!
-    (crate::fs::KDSETKEYCODE as u64, true),
-    (crate::fs::KDSIGACCEPT as u64, true),
-    (libc::TIOCSETD as u64, true),
-    (libc::TIOCSTI as u64, true),
-    (libc::TIOCCONS as u64, true),
-    (libc::TIOCLINUX as u64, true),
-    (crate::fs::FS_IOC_FSGETXATTR as u64, true), // (SAFETY: xattr restrictions)
-    (crate::fs::FS_IOC_FSSETXATTR as u64, true), // ditto!
-    (crate::fs::FS_IOC_SETFLAGS as u64, true),   // To deny immutable & append-only flag changes
+pub const DEFAULT_IOCTL: &[(&str, bool)] = &[
+    ("FIOCLEX", false),
+    ("FIONCLEX", false),
+    ("FIONBIO", false),
+    ("FIONREAD", false),
+    ("FIOASYNC", false),
+    ("FIOQSIZE", false),
+    ("FIFREEZE", true),      // CAP_SYS_ADMIN!
+    ("FITHAW", true),        // ditto!
+    ("FS_IOC_FIEMAP", true), // Leaks on-disk layout.
+    ("FIGETBSZ", true),      // ditto!
+    ("FICLONE", false),
+    ("FICLONERANGE", false),
+    ("FIDEDUPERANGE", false),
+    ("FS_IOC_GETFSUUID", true),      // Leaks on-disk layout.
+    ("FS_IOC_GETFSSYSFSPATH", true), // ditto!
+    ("FIBMAP", true),                // ditto!
+    ("KDSETKEYCODE", true),
+    ("KDSIGACCEPT", true),
+    ("TIOCSETD", true),
+    ("TIOCSTI", true),
+    ("TIOCCONS", true),
+    ("TIOCLINUX", true),
+    ("FS_IOC_FSGETXATTR", true), // (SAFETY: xattr restrictions)
+    ("FS_IOC_FSSETXATTR", true), // ditto!
+    ("FS_IOC_SETFLAGS", true),   // To deny immutable & append-only flag changes
+    ("SECCOMP_IOCTL_NOTIF_RECV", true),
+    ("SECCOMP_IOCTL_NOTIF_SEND", true),
+    ("SECCOMP_IOCTL_NOTIF_ID_VALID", true),
+    ("SECCOMP_IOCTL_NOTIF_ADDFD", true),
+    ("SECCOMP_IOCTL_NOTIF_SET_FLAGS", true),
 ];
 
 // Sandboxing profiles
@@ -579,17 +583,13 @@ pub const PROFILE_LIB: &[&str] = &["include_profile off", "lock:exec"];
 pub const PROFILE_QUIET: &[&str] = &[
     "default/all:filter",
     "default/net,block,mem,pid,force,segvguard,tpe:filter",
-    "filter/read+!proc/name", // for PR_SET_NAME.
 ];
 
 /// Trace sandboxing profile, used by pandora(1).
 pub const PROFILE_TRACE: &[&str] = &[
     "log/level:info",
-    "log/syslog:false",
-    "log/verbose:false",
     "sandbox/ioctl,force:on",
     "default/all,net,force,tpe:warn",
-    "filter/read+!proc/name",
 ];
 
 /// OCI default sandboxing profile, used by syd-oci(1).
@@ -603,7 +603,6 @@ pub const PROFILE_OCI: &[&str] = &[
 /// Enforce profile, used to practically undo `oci` and `trace` profiles at runtime.
 pub const PROFILE_ENFORCE: &[&str] = &[
     "default/all,net,force,tpe:deny",
-    "filter/read-!proc/name",
     "trace/allow_safe_syslog:false",
     "trace/allow_unsafe_nopie:false",
     "lock:exec", // keep esyd.
@@ -650,10 +649,9 @@ pub const PROFILE_NOPIE: &[&str] = &["trace/allow_unsafe_nopie:true"];
 pub const PROFILE_RAND: &[&str] = &["trace/force_rand_fd:true"];
 
 /// Firefox sandboxing profile.
-/// `syd --profile ff`
+/// `syd --profile firefox`
 pub const PROFILE_FIREFOX: &[&str] = &[
-    "filter/read+!proc/name",             // for PR_SET_NAME.
-    "trace/allow_unsafe_filename:true",   // Allow profile directories with curly braces
+    "sandbox/pty:off",                    // firefox(1) is not a console application
     "trace/allow_unsafe_libc:true",       // Allow preloading libmozsandbox.so
     "trace/allow_unsafe_memory:true",     // W^X memory required for JIT
     "trace/allow_unsafe_page_cache:true", // mincore(2) may be needed for DRM
@@ -661,11 +659,14 @@ pub const PROFILE_FIREFOX: &[&str] = &[
 
 /// Readonly sandboxing profile.
 /// `syd --profile ro`
-pub const PROFILE_READONLY: &[&str] = &["deny/wrset+/***"];
+pub const PROFILE_READONLY: &[&str] = &["deny/cpath,wpath+/***"];
 
 /// Container sandboxing profile, activated with:
 /// `syd --profile container`
-pub const PROFILE_CONTAINER: &[&str] = &["unshare/all:true"];
+pub const PROFILE_CONTAINER: &[&str] = &[
+    "unshare/all:true",
+    "allow/read,stat+/proc/uptime", // time-ns aware
+];
 
 /// Immutable container sandboxing profile, activated with:
 /// `syd --profile immutable`
@@ -687,24 +688,28 @@ pub const PROFILE_IMMUTABLE: &[&str] = &[
 /// `syd --profile landlock`.
 pub const PROFILE_LANDLOCK: &[&str] = &[
     "sandbox/lock:on",
-    "allow/lock/read,readdir,exec+/bin",
-    "allow/lock/read,readdir,ioctl+/dev",
-    "allow/lock/read,readdir,exec+/etc",
-    "allow/lock/read,readdir+/lib",
-    "allow/lock/read,readdir,exec+/opt",
+    "allow/lock/rpath,exec+/bin",
+    "allow/lock/rpath,ioctl+/dev",
+    "allow/lock/rpath,exec+/etc",
+    "allow/lock/rpath+/lib",
+    "allow/lock/rpath,exec+/opt",
     // /proc is always allowed.
-    "allow/lock/read,readdir+/run",
-    "allow/lock/read,readdir,exec+/sbin",
-    "allow/lock/read,readdir,exec+/usr",
-    "allow/lock/read,readdir+/var",
+    "allow/lock/rpath+/run",
+    "allow/lock/rpath,exec+/sbin",
+    "allow/lock/rpath,exec+/usr",
+    "allow/lock/rpath+/var",
     "allow/lock/read,write+/dev/full",
     // /dev/null is always allowed.
     "allow/lock/read+/dev/random",
     "allow/lock/read+/dev/urandom",
     "allow/lock/read,write+/dev/zero",
-    "allow/lock/read,readdir,exec,write,create,delete,rename,symlink,truncate,mkdir,rmdir,mkfifo+/dev/shm",
-    "allow/lock/read,readdir,exec,write,create,delete,rename,symlink,truncate,mkdir,rmdir,mkfifo,bind+/tmp",
-    "allow/lock/read,readdir,exec,write,create,delete,rename,symlink,truncate,mkdir,rmdir,mkfifo,bind+/var/tmp",
+    // As of version 3.38.0 we don't allow bind, exec,
+    // and spath access to public directories here.
+    // Paludis profile overrides this as it may be needed
+    // during package testing.
+    "allow/lock/rpath,cpath,wpath,tpath+/dev/shm",
+    "allow/lock/rpath,cpath,wpath,tpath+/tmp",
+    "allow/lock/rpath,cpath,wpath,tpath+/var/tmp",
 ];
 
 /// Paludis sandboxing profile, activated with:
@@ -713,12 +718,11 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "include_profile linux",
     "include_profile tty",
     "default/lock:warn", // set Landlock compat-level to best-effort when stacked with Landlock profile.
-    "allow/lock/read,readdir+/sys", // Relax Landlock to allow /sys when stacked with Landlock profile.
+    "allow/lock/rpath+/sys", // Relax Landlock to allow /sys when stacked with Landlock profile.
     "name/host:localhost", // only effective when combined with unshare/uts:1
     "lock:exec",         // esandbox
-    "log/verbose:false", // verbose logging is intended for malware analysis.
     "mask^",             // do not inherit the default mask-list from linux profile.
-    "sandbox/pty:off",    // Paludis creates a PTY on its own.
+    "sandbox/pty:off",   // Paludis creates a PTY on its own.
     "trace/memory_access:0", // fallback to unsafe proc_pid_mem(5) if cross memory attach fails.
     "trace/allow_safe_bind:true",
     "trace/allow_safe_kcapi:true",
@@ -729,24 +733,26 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "trace/allow_unsafe_exec:true",
     "trace/allow_unsafe_filename:true", // sydbox#118
     "trace/allow_unsafe_libc:true",
-    "trace/allow_unsafe_create:true",     // poppler-data src_install fails without this.
-    "trace/allow_unsafe_hardlinks:true",  // gnulib linkat test fails without this.
+    "trace/allow_unsafe_create:true", // poppler-data src_install fails without this.
+    "trace/allow_unsafe_hardlinks:true", // gnulib linkat test fails without this.
     "trace/allow_unsafe_magiclinks:true", // tests love to access /proc/1.
-    "trace/allow_unsafe_symlinks:true",   // git's tests fail without this.
-    "trace/allow_unsafe_memfd:true",      // executable memory-fds are OK.
+    "trace/allow_unsafe_symlinks:true", // git's tests fail without this.
+    "trace/allow_unsafe_memfd:true",  // executable memory-fds are OK.
     "trace/allow_unsafe_memory:true",
     "trace/allow_unsafe_msgsnd:true",
     "trace/allow_unsafe_nice:true",
     "trace/allow_unsafe_nocookie:true", // syscookies are a mitigation against compromised Syd.
-    "trace/allow_unsafe_nomseal:true", // mseal(2) is a mitigation against compromised Syd.
+    "trace/allow_unsafe_nomseal:true",  // mseal(2) is a mitigation against compromised Syd.
     "trace/allow_unsafe_nopie:true",
     "trace/allow_unsafe_open_path:true",
     "trace/allow_unsafe_page_cache:true",
     "trace/allow_unsafe_perf:true",
+    "trace/allow_unsafe_personality:true", // gawk tests require ADDR_NO_RANDOMIZE.
     "trace/allow_unsafe_ptrace:true",
     "trace/allow_unsafe_prctl:true", // perl tests want to set process name.
     "trace/allow_unsafe_prlimit:true",
-    "trace/allow_unsafe_shm:true",   // perl tests map executable shared memory.
+    "trace/allow_unsafe_proc_status:true", // proc_pid_status(5) hardening is for malware-analysis.
+    "trace/allow_unsafe_shm:true",         // perl tests map executable shared memory.
     "trace/allow_unsafe_sigreturn:true",
     "trace/allow_unsafe_spec_exec:true",
     "trace/allow_unsafe_socket:true",
@@ -759,9 +765,14 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     // See: #208, #212, and #219.
     // no-op unless Landlock is enabled.
     "allow/lock/readdir+/",
+    // Allow wider permissions for public directories.
+    // no-op unless Landlock is enabled.
+    "allow/lock/bind,exec,spath+/dev/shm",
+    "allow/lock/bind,exec,spath+/tmp",
+    "allow/lock/bind,exec,spath+/var/tmp",
     // Allow FS_IOC_SETFLAGS (used by libarchive, cpio, tar etc.)
-    "ioctl/deny-0x40086602", // needed because it's denied by default.
-    "ioctl/allow+0x40086602",
+    "ioctl/deny-FS_IOC_SETFLAGS", // needed because it's denied by default.
+    "ioctl/allow+FS_IOC_SETFLAGS",
     // Turned on by esandbox in metadata phase.
     "sandbox/exec:off",
     // TPE is unnecessary for package builds.
@@ -771,7 +782,7 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     // Defaults for /selinux.
     "allow/stat,chdir,readdir+/selinux",
     "allow/stat,chdir,readdir+/selinux/context",
-    "allow/read,stat,chdir,readdir,write+/selinux/context/**",
+    "allow/rpath,write+/selinux/context/**",
     // Defaults for sysfs.
     "allow/stat,chdir,readdir+/sys",
     "allow/stat,chdir,readdir+/sys/devices",
@@ -784,12 +795,12 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "allow/read,stat+/sys/devices/system/cpu/possible",
     "allow/read,stat+/sys/devices/system/cpu/present",
     "allow/stat,chdir,readdir+/sys/fs",
-    "allow/read,stat,chdir,readdir+/sys/fs/cgroup/***",
+    "allow/rpath+/sys/fs/cgroup/***",
     "allow/stat,chdir,readdir+/sys/kernel",
     "allow/stat,chdir,readdir+/sys/kernel/mm",
-    "allow/read,stat,chdir,readdir+/sys/kernel/mm/transparent_hugepage/***",
+    "allow/rpath+/sys/kernel/mm/transparent_hugepage/***",
     "allow/stat,chdir,readdir+/sys/kernel/security",
-    "allow/read,stat,chdir,readdir+/sys/kernel/security/apparmor/***",
+    "allow/rpath+/sys/kernel/security/apparmor/***",
     // Defaults for procfs.
     "allow/read,stat+/proc/crypto",
     "allow/read,stat+/proc/cpuinfo",
@@ -813,18 +824,23 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     // below have been moved from Linux to Paludis profile.
     "allow/read,stat+/proc/self/maps",
     "allow/read,stat+/proc/self/task/[0-9]*/maps",
+    "allow/read,stat+/proc/mounts", // symlink to self/mounts
     "allow/read,stat+/proc/self/mounts",
     "allow/read,stat+/proc/self/task/[0-9]*/mounts",
-    "allow/read,stat+/proc/mounts", // symlink to self/mounts
-    "allow/read,stat,chdir,readdir+/proc/self/attr/***",
-    "allow/read,stat,chdir,readdir+/proc/self/task/[0-9]*/attr/***",
-    "allow/read,stat,chdir,readdir+/proc/sys",
-    "allow/read,stat,chdir,readdir+/proc/sys/fs/***",
-    "allow/read,stat,chdir,readdir+/proc/sys/kernel/***",
-    "allow/read,stat,chdir,readdir+/proc/sys/vm/***",
+    "allow/read,stat+/proc/self/mountinfo",
+    "allow/read,stat+/proc/self/task/[0-9]*/mountinfo",
+    "allow/rpath+/proc/self/attr/***",
+    "allow/rpath+/proc/self/task/[0-9]*/attr/***",
+    "allow/rpath+/proc/self/cgroup",
+    "allow/rpath+/proc/self/task/[0-9]*/cgroup",
+    "allow/rpath+/proc/sys",
+    "allow/rpath+/proc/sys/fs/***",
+    "allow/rpath+/proc/sys/kernel/***",
+    "allow/rpath+/proc/sys/net/***",
+    "allow/rpath+/proc/sys/vm/***",
     // Allow /var/cache for Paludis builds.
-    "allow/read,stat,chdir,readdir,mktemp+/var/cache/***",
-    "allow/write,create,delete,mkdir,rmdir,mkfifo,rename,symlink,truncate,chown,chgrp,chmod,chattr,utime+/var/cache/**",
+    "allow/rpath,mktemp+/var/cache/***",
+    "allow/cpath,wpath,tpath,fown,fattr,spath+/var/cache/**",
     // Defaults for network sandboxing.
     "allow/net/bind+loopback!0",
     "allow/net/bind+loopback!1024-65535",
@@ -833,10 +849,11 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "allow/lock/bind+1024-65535",
     "allow/lock/connect+1024-65535",
     // Allow interaction with memory file descriptors.
-    "allow/stat,exec,read,write,create,chmod,truncate+/memfd:**",
-    "allow/stat,exec,read,write,create,chmod,truncate+/memfd-hugetlb:**",
+    "allow/create,exec,truncate+!memfd:**",
+    "allow/create,exec,truncate+!memfd-hugetlb:**",
     // Repository directory
-    "allow/read,stat,chdir,readdir+/var/db/paludis/***",
+    "allow/rpath+/var/db",
+    "allow/rpath+/var/db/paludis/***",
 ];
 
 /// User sandboxing profile, activated with:
@@ -846,6 +863,7 @@ pub const PROFILE_PALUDIS: &[&str] = &[
 /// profile.
 pub const PROFILE_USER: &[&str] = &[
     "include_profile linux",
+    "include_profile rand",
     "include_profile tty",
     // dmesg(8)
     "trace/allow_safe_syslog:true",
@@ -854,47 +872,30 @@ pub const PROFILE_USER: &[&str] = &[
     "tpe/user_owned:1",
     "tpe/gid:${SYD_GID}",
     // Enforce strict file modes:
+    // Disallow executable bits to enhance W^X.
     // Disallow s{u,g}id bits on files.
     // Disallow setting group+other bits.
-    "trace/force_umask:7077",
+    "trace/force_umask:7177",
     // /home
-    // 1. We disallow editing dotfiles.
-    // 2. We mark shell-history files as append-only.
-    // 3. We allow read(${HOME}) but not write(${HOME}),
+    // 1. We allow read(${HOME}) but not write(${HOME}),
     //    read|write(${HOME}/**) is ok, i.e. the user can not delete
     //    their home directory under the sandbox which is a nice and
     //    funny protection.
-    "deny/all+/home/**/.*/***",
-    "allow/read,stat,chdir,readdir+/home/**/.*/***",
-    "allow/all+/home/**/.*history*",
-    "append+${HOME}/.*history",
-    "allow/all+/dev/shm/**", // immutable mounts a tmpfs here.
-    "allow/all+/run/**",     // ditto.
-    "allow/all+/var/run/**", // ditto.
-    "allow/read,stat,chdir,readdir+/dev/shm",
-    "allow/read,stat,chdir,readdir+/run",
-    "allow/read,stat,chdir,readdir+/var/run",
-    "allow/read,stat,chdir,readdir+${HOME}/***",
+    // 2. We disallow all dotfiles except history.
+    // 3. We mark shell-history files as append-only.
     "allow/all+${HOME}/**",
+    "allow/rpath+${HOME}/***",
     "allow/net/bind+${HOME}/**",
     "allow/net/connect+${HOME}/**",
     "allow/net/sendfd+${HOME}/**",
+    "deny/all+${HOME}/**/.*/***",
+    "allow/all+${HOME}/**/.*history*",
+    "append+${HOME}/.*history",
     // /run/user/uid
     "allow/all+/run/user/${SYD_UID}/**",
-    "allow/read,stat,chdir,readdir+/run/user/${SYD_UID}/***",
-    "allow/net/connect+/run/user/${SYD_UID}/**",
-    // Allow access to DRI and sound.
-    "allow/read,stat,chdir,readdir+/proc/asound/***",
-    "allow/read,stat,chdir,readdir,write,ioctl+/dev/dri/**",
-    "allow/read,stat,chdir,readdir,write,ioctl+/dev/input/**",
-    "allow/read,stat,chdir,readdir,write,ioctl+/dev/snd/**",
-    // Allow access to HID devices
-    // FIXME: Landlock for hidraw?
-    "allow/read,stat,write,ioctl+/dev/hidraw[0-9]*",
-    // Allow access to syslog.
-    "allow/net/connect+/dev/log",
-    // Allow access to systemd journal.
-    "allow/read,stat,chdir,readdir+/var/log/journal/***",
+    "allow/rpath+/run/user/${SYD_UID}/***",
+    "allow/net/connect+/run/user/${SYD_UID}/***",
+    "allow/net/sendfd+/run/user/${SYD_UID}/***",
 ];
 
 /// Common Linux system profile, used by oci, paludis and user profiles.
@@ -902,14 +903,14 @@ pub const PROFILE_LINUX: &[&str] = &[
     // List root is safe.
     "allow/stat,chdir,readdir+/",
     // Safe defaults for Exec sandboxing
-    "allow/exec,read,stat,chdir,readdir+/bin/***",
-    "allow/exec,read,stat,chdir,readdir+/sbin/***",
-    "allow/exec,read,stat,chdir,readdir+/lib*/***",
-    "allow/exec,read,stat,chdir,readdir+/usr/***",
-    "allow/exec,read,stat,chdir,readdir+/opt/***",
+    "allow/exec,rpath+/bin/***",
+    "allow/exec,rpath+/sbin/***",
+    "allow/exec,rpath+/lib*/***",
+    "allow/exec,rpath+/usr/***",
+    "allow/exec,rpath+/opt/***",
     // SAFETY: Do not leak kernel config/module information.
     // This overrides the lib rule above.
-    "deny/exec,read,stat,chdir,readdir+/lib*/modules/***",
+    "deny/exec,rpath+/lib*/modules/***",
     // Safe defaults for Network sandboxing
     "allow/net/link+route", // allow NETLINK_ROUTE.
     "allow/net/connect+/run/nscd/socket",
@@ -917,10 +918,10 @@ pub const PROFILE_LINUX: &[&str] = &[
     "allow/net/connect+/var/lib/sss/pipes/nss",
     "allow/net/connect+loopback!65535", // getaddrinfo() with AI_ADDRCONFIG on musl.
     // /dev
-    "allow/read,stat,chdir,readdir+/dev",
-    "allow/read,stat,chdir,readdir+/dev/fd",
-    "allow/read,stat,chdir,readdir,mktemp+/dev/shm/***",
-    "allow/write,create,delete,mkdir,rmdir,mkfifo,rename,symlink,truncate,chown,chgrp,chmod,chattr,utime+/dev/shm/**",
+    "allow/rpath+/dev",
+    "allow/rpath+/dev/fd",
+    "allow/rpath,mktemp+/dev/shm/***",
+    "allow/cpath,wpath,tpath,fown,fattr,spath+/dev/shm/**",
     "allow/read,stat,write,ioctl+/dev/full",
     "allow/read,stat,write,ioctl+/dev/zero",
     "allow/read,stat,write,ioctl,truncate+/dev/null",
@@ -936,7 +937,7 @@ pub const PROFILE_LINUX: &[&str] = &[
     // however we do _not_ allow PID traversals,
     // ie identifying fellow processing by readdir'ing `/proc`.
     // Notably, this is allowed in the `paludis` profile.
-    "allow/read,stat,chdir,readdir+/proc",
+    "allow/rpath+/proc",
     // SAFETY: We allow basic stat access (not global but per-process and per-task).
     // trace/allow_unsafe_magiclinks:0 default is another layer of
     // protection against potential malicious activity with proc
@@ -1030,8 +1031,8 @@ pub const PROFILE_LINUX: &[&str] = &[
     r"allow/read,stat,write,truncate,ioctl+/proc/self/fd/socket:\[[0-9]*\]",
     r"allow/read,stat,write,truncate,ioctl+/proc/self/task/[0-9]*/fd/pipe:\[[0-9]*\]",
     r"allow/read,stat,write,truncate,ioctl+/proc/self/task/[0-9]*/fd/socket:\[[0-9]*\]",
-    "allow/read,stat,chdir,readdir+/proc/self/fdinfo/[0-9]*",
-    "allow/read,stat,chdir,readdir+/proc/self/task/[0-9]*/fdinfo/[0-9]*",
+    "allow/rpath+/proc/self/fdinfo/[0-9]*",
+    "allow/rpath+/proc/self/task/[0-9]*/fdinfo/[0-9]*",
     "allow/stat,chdir,readdir+/proc/self/root",
     "allow/stat,chdir,readdir+/proc/self/task/[0-9]*/root",
     // /run
@@ -1040,10 +1041,10 @@ pub const PROFILE_LINUX: &[&str] = &[
     "allow/stat,chdir,readdir+/run/systemd/resolve",
     "allow/read,stat+/run/systemd/resolve/*.conf",
     // /tmp and /var/tmp
-    "allow/read,stat,chdir,readdir,mktemp+/tmp/***",
-    "allow/read,stat,chdir,readdir,mktemp+/var/tmp/***",
-    "allow/write,create,delete,mkdir,rmdir,mkfifo,rename,symlink,truncate,chown,chgrp,chmod,chattr,utime+/tmp/**",
-    "allow/write,create,delete,mkdir,rmdir,mkfifo,rename,symlink,truncate,chown,chgrp,chmod,chattr,utime+/var/tmp/**",
+    "allow/rpath,mktemp+/tmp/***",
+    "allow/rpath,mktemp+/var/tmp/***",
+    "allow/cpath,wpath,tpath,fown,spath+/tmp/**",
+    "allow/cpath,wpath,tpath,fown,fattr,spath+/var/tmp/**",
     // /var
     "allow/stat,chdir,readdir+/var",
     "allow/stat,chdir,readdir,chroot+/var/empty",
@@ -1053,11 +1054,12 @@ pub const PROFILE_LINUX: &[&str] = &[
     "allow/stat,chdir,readdir+/etc",
     "allow/read,stat+/etc/DIR_COLORS",
     "allow/read,stat+/etc/GREP_COLORS",
-    "allow/read,stat,chdir,readdir+/etc/bash*/***",
-    "allow/read,stat,chdir,readdir+/etc/ca-certificates/***",
-    "allow/read,stat,chdir,readdir+/etc/env.d/***",
-    "allow/read,stat,chdir,readdir+/etc/groff/***",
-    "allow/read,stat,chdir,readdir+/etc/ld.so.conf.d/***",
+    "allow/rpath+/etc/bash*/***",
+    "allow/rpath+/etc/alternatives/***",
+    "allow/rpath+/etc/ca-certificates/***",
+    "allow/rpath+/etc/env.d/***",
+    "allow/rpath+/etc/groff/***",
+    "allow/rpath+/etc/ld.so.conf.d/***",
     "allow/read,stat+/etc/environment",
     "allow/read,stat+/etc/ethertypes",
     "allow/read,stat+/etc/group",
@@ -1075,18 +1077,18 @@ pub const PROFILE_LINUX: &[&str] = &[
     "allow/read,stat+/etc/nanorc",
     "allow/read,stat+/etc/**/nsswitch.conf",
     "allow/read,stat+/etc/passwd",
-    "allow/read,stat,chdir,readdir+/etc/profile*/***",
+    "allow/rpath+/etc/profile*/***",
     "allow/read,stat+/etc/services",
     "allow/read,stat+/etc/*-release",
     "allow/read,stat+/etc/protocols",
     "allow/read,stat+/etc/resolv.conf",
-    "allow/read,stat,chdir,readdir+/etc/skel/***",
+    "allow/rpath+/etc/skel/***",
     "allow/stat,chdir,readdir+/etc/ssl",
-    "allow/read,stat,chdir,readdir+/etc/ssl/certs/***",
-    "allow/read,stat,chdir,readdir+/etc/ssl/misc/***",
+    "allow/rpath+/etc/ssl/certs/***",
+    "allow/rpath+/etc/ssl/misc/***",
     "allow/read,stat+/etc/ssl/openssl.cnf",
-    "allow/read,stat,chdir,readdir+/etc/terminfo/***",
-    "allow/read,stat,chdir,readdir+/etc/zsh/***",
+    "allow/rpath+/etc/terminfo/***",
+    "allow/rpath+/etc/zsh/***",
     // /home
     //
     // Do _not_ allow readdir which allows enumerating other users!
@@ -1103,9 +1105,11 @@ pub const PROFILE_LINUX: &[&str] = &[
 /// paludis and user profiles.
 /// Syd sets the environment variable TTY before parsing this profile.
 pub const PROFILE_TTY: &[&str] = &[
+    // Allow safe ioctl requests without path check.
+    "include_profile tty_native",
     // TTY lock rules to be used with Landlock.
     "allow/lock/read,write,ioctl+/dev/ptmx",
-    "allow/lock/read,readdir,write,ioctl+/dev/pts",
+    "allow/lock/rpath,write,ioctl+/dev/pts",
     "allow/lock/read,write,ioctl+/dev/tty",
     "allow/lock/read,write,ioctl+${SYD_TTY}",
     // TTY seccomp rules to act as the second layer.
@@ -1114,35 +1118,37 @@ pub const PROFILE_TTY: &[&str] = &[
     "allow/read,stat,write,ioctl+/dev/pts/[0-9]*",
     "allow/read,stat,write,ioctl+/dev/tty",
     "allow/read,stat,write,ioctl+${SYD_TTY}",
-    // PTY handling
-    "ioctl/allow+0x540f",     // TIOCGPGRP: tcgetpgrp(3)
-    "ioctl/allow+0x5410",     // TIOCSPGRP: tcsetpgrp(3)
-    "ioctl/allow+0x5429",     // TIOCGSID: tcgetsid(3)
-    "ioctl/allow+0x5401",     // TCGETS: isatty(3), tcgetattr(3) etc.
-    "ioctl/allow+0x802c542a", // TCGETS2: ditto.
-    "ioctl/allow+0x5432",     // TCGETX
-    "ioctl/allow+0x5433",     // TCSETX
-    "ioctl/allow+0x5434",     // TCSETXF
-    "ioctl/allow+0x5435",     // TCSETXW
-    "ioctl/allow+0x5456",     // TIOCGLCKTRMIOS
-    "ioctl/allow+0x5457",     // TIOCSLCKTRMIOS
-    "ioctl/allow+0x5413",     // TIOCGWINSZ
-    "ioctl/allow+0x5414",     // TIOCSWINSZ
-    "ioctl/allow+0x80045432", // TIOCGDEV
-    "ioctl/allow+0x80045440", // TIOCGEXCL
-    "ioctl/allow+0x80045438", // TIOCGPKT
-    "ioctl/allow+0x80045439", // TIOCGPTLCK
-    "ioctl/allow+0x80045430", // TIOCGPTN
-    "ioctl/allow+0x5441",     // TIOCGPTPEER
-    "ioctl/allow+0x540e",     // TIOCSCTTY
-    "ioctl/allow+0x40045436", // TIOCSIG
-    "ioctl/allow+0x5437",     // TIOCVHANGUP
-    // Pseudoterminal ioctls
-    "ioctl/allow+0x80045438", // TIOCGPKT
-    "ioctl/allow+0x80045439", // TIOCGPTLCK
-    "ioctl/allow+0x5441",     // TIOCGPTPEER
-    "ioctl/allow+0x5420",     // TIOCPKT
-    "ioctl/allow+0x40045431", // TIOCSPTLCK
+];
+
+/// Profile to allowlist TTY-native ioctls without path check.
+pub const PROFILE_TTY_NATIVE: &[&str] = &[
+    "ioctl/allow+TIOCGPGRP",
+    "ioctl/allow+TIOCSPGRP",
+    "ioctl/allow+TIOCGSID",
+    "ioctl/allow+TCGETS",
+    "ioctl/allow+TCGETS2",
+    "ioctl/allow+TCGETX",
+    "ioctl/allow+TCSETX",
+    "ioctl/allow+TCSETXF",
+    "ioctl/allow+TCSETXW",
+    "ioctl/allow+TIOCGLCKTRMIOS",
+    "ioctl/allow+TIOCSLCKTRMIOS",
+    "ioctl/allow+TIOCGWINSZ",
+    "ioctl/allow+TIOCSWINSZ",
+    "ioctl/allow+TIOCGDEV",
+    "ioctl/allow+TIOCGEXCL",
+    "ioctl/allow+TIOCGPKT",
+    "ioctl/allow+TIOCGPTLCK",
+    "ioctl/allow+TIOCGPTN",
+    "ioctl/allow+TIOCGPTPEER",
+    "ioctl/allow+TIOCSCTTY",
+    "ioctl/allow+TIOCSIG",
+    "ioctl/allow+TIOCVHANGUP",
+    "ioctl/allow+TIOCGPKT",
+    "ioctl/allow+TIOCGPTLCK",
+    "ioctl/allow+TIOCGPTPEER",
+    "ioctl/allow+TIOCPKT",
+    "ioctl/allow+TIOCSPTLCK",
 ];
 
 /// Profile to allowlist KVM ioctls without path check.
@@ -1150,48 +1156,48 @@ pub const PROFILE_TTY: &[&str] = &[
 pub const PROFILE_KVM: &[&str] = &[
     "allow/read,stat,write,ioctl+/dev/kvm",
     "include_profile kvm_native",
-    "ioctl/allow+0x0000ae00", // KVM_GET_API_VERSION
-    "ioctl/allow+0xae01",     // KVM_CREATE_VM
-    "ioctl/allow+0xae04",     // KVM_GET_VCPU_MMAP_SIZE
-    "ioctl/allow+0xae41",     // KVM_CREATE_VCPU
-    "ioctl/allow+0x4010ae42", // KVM_GET_DIRTY_LOG
-    "ioctl/allow+0xae47",     // KVM_SET_TSS_ADDR
-    "ioctl/allow+0xae80",     // KVM_RUN
-    "ioctl/allow+0xae9a",     // KVM_NMI
-    "ioctl/allow+0xae03",     // KVM_CHECK_EXTENSION
-    "ioctl/allow+0xaea3",     // KVM_GET_TSC_KHZ
-    "ioctl/allow+0xaea2",     // KVM_SET_TSC_KHZ
-    "ioctl/allow+0x4004ae86", // KVM_INTERRUPT
-    "ioctl/allow+0x4008ae89", // KVM_SET_MSRS
-    "ioctl/allow+0x4020ae46", // KVM_SET_USER_MEMORY_REGION
-    "ioctl/allow+0x4090ae82", // KVM_SET_REGS
-    "ioctl/allow+0x4138ae84", // KVM_SET_SREGS
-    "ioctl/allow+0xc008ae88", // KVM_GET_MSRS
-    "ioctl/allow+0x8090ae81", // KVM_GET_REGS
-    "ioctl/allow+0x8138ae83", // KVM_GET_SREGS
-    "ioctl/allow+0xc008ae05", // KVM_GET_SUPPORTED_CPUID
-    "ioctl/allow+0xc008ae09", // KVM_GET_EMULATED_CPUID
-    "ioctl/allow+0x4008ae90", // KVM_SET_CPUID2
-    "ioctl/allow+0x4004ae8b", // KVM_SET_SIGNAL_MASK
-    "ioctl/allow+0x8040ae9f", // KVM_GET_VCPU_EVENTS
-    "ioctl/allow+0x4040aea0", // KVM_SET_VCPU_EVENTS
-    "ioctl/allow+0x4018aee1", // KVM_SET_DEVICE_ATTR
-    "ioctl/allow+0x4008ae48", // KVM_SET_IDENTITY_MAP_ADDR
-    "ioctl/allow+0x0000ae60", // KVM_CREATE_IRQCHIP
-    "ioctl/allow+0x4008ae61", // KVM_IRQ_LINE
-    "ioctl/allow+0x4010ae67", // KVM_REGISTER_COALESCED_MMIO
-    "ioctl/allow+0x4010ae68", // KVM_UNREGISTER_COALESCED_MMIO
-    "ioctl/allow+0x4030ae6a", // KVM_SET_GSI_ROUTING
-    "ioctl/allow+0x4020ae76", // KVM_IRQFD
-    "ioctl/allow+0x4040ae79", // KVM_IOEVENTFD
-    "ioctl/allow+0x8004ae98", // KVM_GET_MP_STATE
-    "ioctl/allow+0x4004ae99", // KVM_SET_MP_STATE
-    "ioctl/allow+0x4200aea5", // KVM_SIGNAL_MSI
-    "ioctl/allow+0x404ae09b", // KVM_SET_GUEST_DEBUG
-    "ioctl/allow+0xc00caee0", // KVM_CREATE_DEVICE
-    "ioctl/allow+0x4018aee1", // KVM_SET_DEVICE_ATTR
-    "ioctl/allow+0x8018aee2", // KVM_GET_DEVICE_ATTR
-    "ioctl/allow+0x4018aee3", // KVM_HAS_DEVICE_ATTR
+    "ioctl/allow+KVM_GET_API_VERSION",
+    "ioctl/allow+KVM_CREATE_VM",
+    "ioctl/allow+KVM_GET_VCPU_MMAP_SIZE",
+    "ioctl/allow+KVM_CREATE_VCPU",
+    "ioctl/allow+KVM_GET_DIRTY_LOG",
+    "ioctl/allow+KVM_SET_TSS_ADDR",
+    "ioctl/allow+KVM_RUN",
+    "ioctl/allow+KVM_NMI",
+    "ioctl/allow+KVM_CHECK_EXTENSION",
+    "ioctl/allow+KVM_GET_TSC_KHZ",
+    "ioctl/allow+KVM_SET_TSC_KHZ",
+    "ioctl/allow+KVM_INTERRUPT",
+    "ioctl/allow+KVM_SET_MSRS",
+    "ioctl/allow+KVM_SET_USER_MEMORY_REGION",
+    "ioctl/allow+KVM_SET_REGS",
+    "ioctl/allow+KVM_SET_SREGS",
+    "ioctl/allow+KVM_GET_MSRS",
+    "ioctl/allow+KVM_GET_REGS",
+    "ioctl/allow+KVM_GET_SREGS",
+    "ioctl/allow+KVM_GET_SUPPORTED_CPUID",
+    "ioctl/allow+KVM_GET_EMULATED_CPUID",
+    "ioctl/allow+KVM_SET_CPUID2",
+    "ioctl/allow+KVM_SET_SIGNAL_MASK",
+    "ioctl/allow+KVM_GET_VCPU_EVENTS",
+    "ioctl/allow+KVM_SET_VCPU_EVENTS",
+    "ioctl/allow+KVM_SET_DEVICE_ATTR",
+    "ioctl/allow+KVM_SET_IDENTITY_MAP_ADDR",
+    "ioctl/allow+KVM_CREATE_IRQCHIP",
+    "ioctl/allow+KVM_IRQ_LINE",
+    "ioctl/allow+KVM_REGISTER_COALESCED_MMIO",
+    "ioctl/allow+KVM_UNREGISTER_COALESCED_MMIO",
+    "ioctl/allow+KVM_SET_GSI_ROUTING",
+    "ioctl/allow+KVM_IRQFD",
+    "ioctl/allow+KVM_IOEVENTFD",
+    "ioctl/allow+KVM_GET_MP_STATE",
+    "ioctl/allow+KVM_SET_MP_STATE",
+    "ioctl/allow+KVM_SIGNAL_MSI",
+    "ioctl/allow+KVM_SET_GUEST_DEBUG",
+    "ioctl/allow+KVM_CREATE_DEVICE",
+    "ioctl/allow+KVM_SET_DEVICE_ATTR",
+    "ioctl/allow+KVM_GET_DEVICE_ATTR",
+    "ioctl/allow+KVM_HAS_DEVICE_ATTR",
 ];
 
 /// Profile to allowlist KVM arch-native ioctls without path check.
@@ -1815,7 +1821,7 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     //"pidfd_open", restrictions applied, see setup_seccomp.
     "pidfd_send_signal",
     "pipe",
-    "pipe2",
+    //"pipe2", restrictions applied, see setup_seccomp.
     //pivot_root,
     //"pkey_alloc", restrictions applied, see setup_seccomp.
     //"pkey_free", ditto
@@ -1838,7 +1844,7 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     //"putpmsg",
     "pwrite64",
     "pwritev",
-    "pwritev2",
+    //"pwritev2", // restrictions applied see setup_seccomp.
     //query_module,
     //quotactl
     //quotactl_fd
@@ -2069,7 +2075,7 @@ pub const EMU_SYSCALLS: &[&str] = &[
     "pidfd_getfd",
     "pidfd_open",
     "pidfd_send_signal",
-    "pipe2",
+    //"pipe2", the flag O_NOTIFICATION_PIPE is disallowed.
     "prctl",
     "process_mrelease", // Used by kill action when signal is SIGKILL.
     "process_vm_readv",
@@ -2083,7 +2089,7 @@ pub const EMU_SYSCALLS: &[&str] = &[
     "rt_sigtimedwait",        // Used by interrupt handling.
     "rt_sigtimedwait_time64", // ditto.
     "sched_getaffinity",
-    "sched_setaffinity",
+    //"sched_setaffinity", we use sync seccomp rather than CPU pinning as of 3.37.6.
     "sched_yield",
     "send",
     "sendmsg",
@@ -2098,6 +2104,10 @@ pub const EMU_SYSCALLS: &[&str] = &[
     "tee",
     "tgkill",
     "tkill",
+    "timer_create",    // needed to unblock idle helper threads.
+    "timer_delete",    // ditto.
+    "timer_settime",   // ditto.
+    "timer_settime64", // ditto.
     //"truncate", protected with syscall argument cookies.
     //"truncate64", ditto.
     "umask",
@@ -2122,6 +2132,9 @@ pub const EMU_SYSCALLS: &[&str] = &[
     "statx",
     //"unlinkat", protected with syscall argument cookies.
     "utimensat",
+    // used by logging and proc_pid_status(5) masking.
+    "write",
+    "writev",
 ];
 
 /// Fcntl operations allowed for emulator threads.
@@ -2132,6 +2145,7 @@ pub const EMU_FCNTL_OPS: &[u64] = &[
     libc::F_SETFL as u64,
     libc::F_OFD_SETLK as u64,
     libc::F_OFD_SETLKW as u64,
+    libc::F_ADD_SEALS as u64,
 ];
 
 /// System calls allowed for Interrupter thread.
@@ -2219,7 +2233,7 @@ pub const AES_SYSCALLS: &[&str] = &[
     "clone",
     "clone3",
     "close",
-    "connect",
+    "connect", // TODO: protect with syscall argument cookies.
     "exit",
     //"fcntl", See: AES_FCNTL_OPS for allowed fcntl operations.
     //"fcntl64",
@@ -2244,7 +2258,7 @@ pub const AES_SYSCALLS: &[&str] = &[
     "munlock",
     "munmap",
     "nanosleep",
-    "pipe2",
+    //"pipe2", the flag O_NOTIFICATION_PIPE is disallowed.
     "prctl",
     "recvmsg",
     "restart_syscall",
@@ -2390,6 +2404,15 @@ pub static HAVE_MFD_NOEXEC_SEAL: Lazy<bool> = Lazy::new(|| {
     major > 6 || (major == 6 && minor >= 3)
 });
 
+/// Returns MFD_NOEXEC_SEAL on Linux>=6.3 and MFD_ALLOW_SEALING on older Linux.
+pub static SAFE_MFD_FLAGS: Lazy<libc::c_uint> = Lazy::new(|| {
+    if *HAVE_MFD_NOEXEC_SEAL {
+        libc::MFD_NOEXEC_SEAL
+    } else {
+        libc::MFD_ALLOW_SEALING
+    }
+});
+
 /// Check for PIDFD_THREAD support (Linux-6.9 or newer).
 pub static HAVE_PIDFD_THREAD: Lazy<bool> = Lazy::new(|| {
     let (major, minor) = *KERNEL_VERSION;
@@ -2435,6 +2458,12 @@ pub static HAVE_PROC_PID_FD_STAT_SIZE: Lazy<bool> = Lazy::new(|| {
     major > 6 || (major == 6 && minor >= 2)
 });
 
+/// Check for RWF_NOAPPEND support for pwritev2(2) (Linux-6.9 or newer).
+pub static HAVE_RWF_NOAPPEND: Lazy<bool> = Lazy::new(|| {
+    let (major, minor) = *KERNEL_VERSION;
+    major > 6 || (major == 6 && minor >= 9)
+});
+
 //
 // Below is internal territory, you have been warned.
 
@@ -2442,21 +2471,23 @@ pub static HAVE_PROC_PID_FD_STAT_SIZE: Lazy<bool> = Lazy::new(|| {
 pub(crate) const SHM_UNSAFE_MASK: u64 = 0o177;
 
 /// Name prefix for memory file descriptors.
-pub const MFD_NAME_PREFIX: &[u8] = b"/memfd:";
+/// This is used in access control.
+pub const MFD_NAME_PREFIX: &[u8] = b"!memfd:";
 
 /// Name prefix for hugetlbfs memory file descriptors.
-pub const MFD_NAME_PREFIX_HUGETLB: &[u8] = b"/memfd-hugetlb:";
+/// This is used in access control.
+pub const MFD_HUGETLB_NAME_PREFIX: &[u8] = b"!memfd-hugetlb:";
 
 /// BIGTCP maximum for loopback:
 ///
 /// https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=d6f938ce52f9adb23f4c31cc371654a5f18ff328
 pub const LOOPBACK_BIGTCP_MAX: u32 = 0x30d40;
 
-// Determine minimum allowed mmap address by reading `/proc/sys/vm/mmap_min_addr`.
-//
-// Enforce OpenBSD's hard-coded minimum mmap address by capping at PAGE_SIZE;
-// See: https://github.com/openbsd/src/blob/94a7e27d04f23871848a8f225ef286e84969197a/sys/uvm/uvm_map.c#L235
-pub(crate) static MMAP_MIN_ADDR: Lazy<u64> =
+/// Determine minimum allowed mmap(2) address by reading `/proc/sys/vm/mmap_min_addr`.
+///
+/// Enforce OpenBSD's hard-coded minimum mmap(2) address by capping at PAGE_SIZE;
+/// See: https://github.com/openbsd/src/blob/94a7e27d04f23871848a8f225ef286e84969197a/sys/uvm/uvm_map.c#L235
+pub static MMAP_MIN_ADDR: Lazy<u64> =
     Lazy::new(|| proc_mmap_min_addr().unwrap_or(*PAGE_SIZE).max(*PAGE_SIZE));
 
 // Trace data for ptrace(2) hooked system calls.
@@ -2718,7 +2749,9 @@ pub fn proc_init() -> Result<(), Errno> {
         .map(|stx| (stx.stx_mode & !(libc::S_IFMT as u16), stx.stx_mnt_id))?;
 
     // SAFETY: Validate what we've opened is indeed `/dev/null`.
-    crate::validate_dev_null(fd_null.as_raw_fd())?;
+    if !is_dev_null(&fd_null).unwrap_or(false) {
+        return Err(Errno::ENODEV);
+    }
 
     // SAFETY: To make this file descriptor harder to spot by an
     // attacker we duplicate it to a random fd number.
