@@ -6,27 +6,25 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
-use std::{
-    fs::File,
-    io::{Seek, SeekFrom},
-};
+use std::{fs::File, io::Seek, os::fd::AsRawFd};
 
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
-    fcntl::AtFlags,
+    fcntl::{AtFlags, OFlag},
     sys::signal::{kill, Signal},
     unistd::Pid,
     NixPath,
 };
 
 use crate::{
-    config::{PTRACE_DATA_EXECVE, PTRACE_DATA_EXECVEAT},
+    config::{PROC_FILE, PTRACE_DATA_EXECVE, PTRACE_DATA_EXECVEAT},
     elf::{ElfError, ElfFileType, ElfType, ExecutableFile, LinkingType},
     err2no, error,
-    fs::{is_executable, set_nonblock, FileType, FsFlags, MaybeFd, AT_EXECVE_CHECK},
+    fs::{is_executable, safe_open_magicsym, FileType, FsFlags, MaybeFd, AT_EXECVE_CHECK},
     hook::{RemoteProcess, SysArg, SysFlags, UNotifyEventRequest},
     kernel::{sandbox_path, to_atflags},
+    path::XPathBuf,
     ptrace::ptrace_syscall_info,
     sandbox::{Action, Capability, IntegrityError, SandboxGuard},
     warn,
@@ -52,7 +50,7 @@ pub(crate) fn sysenter_exec(
             "execve",
             SysArg {
                 path: Some(0),
-                fsflags: FsFlags::MUST_PATH | FsFlags::WANT_READ,
+                fsflags: FsFlags::MUST_PATH,
                 ..Default::default()
             },
             false,
@@ -64,7 +62,7 @@ pub(crate) fn sysenter_exec(
                 AtFlags::AT_SYMLINK_NOFOLLOW | AtFlags::AT_EMPTY_PATH | AT_EXECVE_CHECK,
             )?;
 
-            let mut fsflags = FsFlags::MUST_PATH | FsFlags::WANT_READ;
+            let mut fsflags = FsFlags::MUST_PATH;
             if flags.contains(AtFlags::AT_SYMLINK_NOFOLLOW) {
                 fsflags.insert(FsFlags::NO_FOLLOW_LAST);
             }
@@ -128,10 +126,14 @@ pub(crate) fn sysenter_exec(
     // SAFETY: Return EACCES without any more processing
     // if the file is not a regular file or a memory fd.
     // Mfd check depends on trace/allow_unsafe_memfd option.
+    //
+    // Note, attempting to execute directories on Linux
+    // return EACCES, not EISDIR like the manual page
+    // claims. GNU make has a test checking this errno.
     match path.typ.as_ref() {
         Some(FileType::Reg) => {}
-        Some(FileType::Mfd) if sandbox.allow_unsafe_memfd() => {}
-        Some(FileType::Dir) => return Err(Errno::EISDIR),
+        Some(FileType::Mfd) if sandbox.flags.allow_unsafe_memfd() => {}
+        //Some(FileType::Dir) => return Err(Errno::EISDIR),
         _ => return Err(Errno::EACCES),
     };
 
@@ -211,50 +213,34 @@ pub(crate) fn sysenter_exec(
         }
     }
 
+    // Open a new instance of the file for read.
+    // We prefer this over WANT_READ to ensure
+    // we do not share OFD with the sandbox process.
     assert!(
         path.base.is_empty(),
         "BUG: MUST_PATH returned a directory for exec, report a bug!"
     );
     let mut file = match path.dir.take() {
         Some(MaybeFd::Owned(fd)) => {
-            set_nonblock(&fd, false)?;
-            File::from(fd)
+            let pfd = XPathBuf::from_self_fd(fd.as_raw_fd());
+            safe_open_magicsym(PROC_FILE(), &pfd, OFlag::O_RDONLY).map(File::from)?
         }
         _ => return Err(Errno::ENOEXEC),
     };
 
     // Parse ELF as necessary for restrictions.
-    let deny_script = sandbox.deny_script();
-    let restrict_32 = sandbox.deny_elf32();
-    let restrict_dyn = sandbox.deny_elf_dynamic();
-    let restrict_sta = sandbox.deny_elf_static();
-    let restrict_pie = !sandbox.allow_unsafe_nopie();
-    let restrict_xs = !sandbox.allow_unsafe_stack();
+    let deny_script = sandbox.flags.deny_script();
+    let restrict_32 = sandbox.flags.deny_elf32();
+    let restrict_dyn = sandbox.flags.deny_elf_dynamic();
+    let restrict_sta = sandbox.flags.deny_elf_static();
+    let restrict_pie = !sandbox.flags.allow_unsafe_nopie();
+    let restrict_xs = !sandbox.flags.allow_unsafe_stack();
 
     // Shared library execution depends on trace/allow_unsafe_exec:1.
-    let restrict_ldd = !sandbox.allow_unsafe_exec();
-
+    let restrict_ldd = !sandbox.flags.allow_unsafe_exec();
     let check_linking = restrict_ldd || restrict_dyn || restrict_sta || restrict_pie || restrict_xs;
 
-    // Ensure the file offset is maintained,
-    // as the file might be sharing the OFD
-    // with the sandbox process.
-    //
-    // SAFETY: Deny non-seekable files such as FIFOs with EACCES.
-    let offset = file.stream_position().or(Err(Errno::EACCES))?;
-
-    let result = (|| -> Result<ExecutableFile, ElfError> {
-        // Parse ELF and reset the file offset.
-        if offset != 0 {
-            file.rewind().map_err(ElfError::IoError)?;
-        }
-        let result = ExecutableFile::parse(&file, check_linking);
-        file.seek(SeekFrom::Start(offset))
-            .map_err(ElfError::IoError)?;
-        result
-    })();
-
-    let exe = match result {
+    let exe = match ExecutableFile::parse(&file, check_linking) {
         Ok(exe) => exe,
         Err(ElfError::IoError(err)) => {
             let errno = err2no(&err);
@@ -392,14 +378,10 @@ pub(crate) fn sysenter_exec(
 
     // Check for Force sandboxing.
     if sandbox.enabled(Capability::CAP_FORCE) {
-        // Calculate checksum and reset the file offset.
-        if offset != 0 {
-            file.rewind().or(Err(Errno::EACCES))?;
-        }
-        let result = sandbox.check_force2(path.abs(), &mut file);
-        file.seek(SeekFrom::Start(offset)).or(Err(Errno::EACCES))?;
+        // Reset the file offset and calculate checksum.
+        file.rewind().or(Err(Errno::EACCES))?;
 
-        match result {
+        match sandbox.check_force2(path.abs(), &mut file) {
             Ok(Action::Allow) => {}
             Ok(Action::Warn) => {
                 warn!("ctx": "verify_elf", "act": Action::Warn,
@@ -553,7 +535,7 @@ fn syscall_exec_handler(
         // and only the file executability check (aka AT_EXECVE_CHECK)
         // will be performed.
         let sandbox = request.get_sandbox();
-        let restrict_memfd = !sandbox.allow_unsafe_memfd();
+        let restrict_memfd = !sandbox.flags.allow_unsafe_memfd();
 
         // Read remote path.
         let (path, _) = request.read_path(&sandbox, arg, false)?;
@@ -588,10 +570,14 @@ fn syscall_exec_handler(
         // SAFETY: Return EACCES without any more processing
         // if the file is not a regular file or a memory fd.
         // Mfd check depends on trace/allow_unsafe_memfd option.
+        //
+        // Note, attempting to execute directories on Linux
+        // return EACCES, not EISDIR like the manual page
+        // claims. GNU make has a test checking this errno.
         match path.typ.as_ref() {
             Some(FileType::Reg) => {}
             Some(FileType::Mfd) if !restrict_memfd => {}
-            Some(FileType::Dir) => return Err(Errno::EISDIR),
+            //Some(FileType::Dir) => return Err(Errno::EISDIR),
             _ => return Err(Errno::EACCES),
         };
 
