@@ -93,9 +93,6 @@ pub(crate) mod shm;
 /// {,rt_}sigaction(2) handler
 pub(crate) mod sigaction;
 
-/// {,rt_}sigreturn(2) handler
-pub(crate) mod sigreturn;
-
 /// Signal syscall handlers
 pub(crate) mod signal;
 
@@ -146,9 +143,10 @@ use nix::{
 use crate::{
     fs::{to_valid_fd, CanonicalPath, FileInfo, FileType},
     hook::{PathArgs, SysArg, UNotifyEventRequest},
+    ioctl::{ioctl_names_get, Ioctl},
     notice,
     path::{XPath, XPathBuf},
-    sandbox::{Action, Capability, SandboxGuard},
+    sandbox::{Action, Capability, Sandbox, SandboxGuard},
     warn,
 };
 
@@ -157,7 +155,7 @@ use crate::{
 #[allow(clippy::too_many_arguments)]
 pub(crate) fn sandbox_path(
     request: Option<&UNotifyEventRequest>,
-    sandbox: &SandboxGuard,
+    sandbox: &Sandbox,
     pid: Pid,
     path: &XPath,
     caps: Capability,
@@ -174,7 +172,7 @@ pub(crate) fn sandbox_path(
     let mut caps = sandbox.getcaps(caps);
     let stat = sandbox.enabled(Capability::CAP_STAT);
     if caps.is_empty() && (!hide || !stat) {
-        return if caps_old.intersects(Capability::CAP_WRSET) && sandbox.is_append(path) {
+        return if caps_old.can_write() && sandbox.is_append(path) {
             // SAFETY: Protect append-only paths against writes.
             // We use ECANCELED which will result in a no-op.
             Err(Errno::ECANCELED)
@@ -252,18 +250,29 @@ pub(crate) fn sandbox_path(
         };
 
         if let Some(request) = request {
+            let arch = request.scmpreq.data.arch;
             let args = request.scmpreq.data.args;
+
+            // Enrich log data with ioctl(2) request names.
+            let ctl = if caps.contains(Capability::CAP_IOCTL) {
+                ioctl_names_get(args[1] as Ioctl, arch)
+            } else {
+                None
+            };
+
             if sandbox.verbose {
                 if is_warn {
                     warn!("ctx": "access", "cap": caps, "act": action,
-                        "sys": syscall_name, "path": &path, "args": args,
+                        "sys": syscall_name, "ctl": ctl,
+                        "path": &path, "args": args,
                         "tip": format!("configure `allow/{}+{}'",
                             caps.to_string().to_ascii_lowercase(),
                             path),
                         "req": request);
                 } else {
                     notice!("ctx": "access", "cap": caps, "act": action,
-                        "sys": syscall_name, "path": &path, "args": args,
+                        "sys": syscall_name, "ctl": ctl,
+                        "path": &path, "args": args,
                         "tip": format!("configure `allow/{}+{}'",
                             caps.to_string().to_ascii_lowercase(),
                             path),
@@ -271,14 +280,16 @@ pub(crate) fn sandbox_path(
                 }
             } else if is_warn {
                 warn!("ctx": "access", "cap": caps, "act": action,
-                    "sys": syscall_name, "path": &path, "args": args,
+                    "sys": syscall_name, "ctl": ctl,
+                    "path": &path, "args": args,
                     "tip": format!("configure `allow/{}+{}'",
                         caps.to_string().to_ascii_lowercase(),
                         path),
                     "pid": request.scmpreq.pid);
             } else {
                 notice!("ctx": "access", "cap": caps, "act": action,
-                    "sys": syscall_name, "path": &path, "args": args,
+                    "sys": syscall_name, "ctl": ctl,
+                    "path": &path, "args": args,
                     "tip": format!("configure `allow/{}+{}'",
                         caps.to_string().to_ascii_lowercase(),
                         path),
@@ -303,7 +314,7 @@ pub(crate) fn sandbox_path(
 
     match action {
         Action::Allow | Action::Warn => {
-            if caps.intersects(Capability::CAP_WRSET) && sandbox.is_append(&path) {
+            if caps.can_write() && sandbox.is_append(&path) {
                 // SAFETY: Protect append-only paths against writes.
                 // We use ECANCELED which will result in a no-op.
                 Err(Errno::ECANCELED)
@@ -453,13 +464,12 @@ where
                         // SAFETY: Only regular files are encrypted.
                         CanonicalPath::new(path, FileType::Reg, arg.fsflags)?
                     } else {
-                        CanonicalPath::new_fd(fd.into(), req.pid(), dirfd)?
+                        CanonicalPath::new_fd(fd.into(), req.pid())?
                     };
 
                     paths[idx] = Some(path);
                 } else {
-                    let path =
-                        CanonicalPath::new_fd(libc::AT_FDCWD.into(), req.pid(), libc::AT_FDCWD)?;
+                    let path = CanonicalPath::new_fd(libc::AT_FDCWD.into(), req.pid())?;
 
                     paths[idx] = Some(path);
                 }
@@ -545,7 +555,7 @@ where
                 }
 
                 if let Some(file_type) = &path.typ {
-                    if !matches!(file_type, FileType::Dir | FileType::MagicLnk(_, _))
+                    if !matches!(file_type, FileType::Dir | FileType::MagicLnk(_))
                         && path.abs().last() == Some(b'/')
                     {
                         return Err(Errno::ENOTDIR);
