@@ -9,6 +9,9 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
+// SAFETY: This module has been liberated from unsafe code!
+#![forbid(unsafe_code)]
+
 // Last sync with pentacle:
 // Version 1.0.0
 // Commit:e606ab250e6655865bb93a6d98157093f2eb455f
@@ -29,7 +32,7 @@ use nix::{
         c_int, c_uint, F_SEAL_FUTURE_WRITE, F_SEAL_GROW, F_SEAL_SEAL, F_SEAL_SHRINK, F_SEAL_WRITE,
         MFD_ALLOW_SEALING, MFD_CLOEXEC, MFD_EXEC, MFD_NOEXEC_SEAL,
     },
-    sys::memfd::MemFdCreateFlag,
+    sys::memfd::MFdFlags,
 };
 
 use crate::err::err2no;
@@ -468,19 +471,15 @@ impl SealOptions {
 }
 
 fn memfd_create(name: &CStr, flags: c_uint) -> Result<File, Errno> {
-    nix::sys::memfd::memfd_create(name, MemFdCreateFlag::from_bits_retain(flags)).map(File::from)
+    nix::sys::memfd::memfd_create(name, MFdFlags::from_bits_retain(flags)).map(File::from)
 }
 
 fn fcntl_get_seals(file: &File) -> Result<c_int, Errno> {
-    fcntl(file.as_raw_fd(), FcntlArg::F_GET_SEALS)
+    fcntl(file, FcntlArg::F_GET_SEALS)
 }
 
 fn fcntl_add_seals(file: &File, arg: c_int) -> Result<(), Errno> {
-    fcntl(
-        file.as_raw_fd(),
-        FcntlArg::F_ADD_SEALS(SealFlag::from_bits_retain(arg)),
-    )
-    .map(drop)
+    fcntl(file, FcntlArg::F_ADD_SEALS(SealFlag::from_bits_retain(arg))).map(drop)
 }
 
 #[cfg(test)]
