@@ -26,7 +26,7 @@ use crate::{
     compat::{waitid, WaitStatus},
     config::CHLD_STACK_SIZE,
     fs::{duprand, pidfd_getfd, retry_on_eintr},
-    unshare::{child, config::Config, Child, Command, Executable},
+    unshare::{child, config::Config, Child, Command, Executable, IoctlMap},
 };
 
 type ChildPreExecFunc = Box<dyn Fn() -> Result<(), Errno>>;
@@ -43,6 +43,7 @@ pub struct ChildInfo<'a> {
     pub cfg: Config,
     pub pre_exec: Option<ChildPreExecFunc>,
     pub pty_fd: Option<RawFd>,
+    pub ioctl_denylist: Option<IoctlMap>,
     pub seccomp_filter: Option<ScmpFilterContext>,
     pub seccomp_pipefd: PipePair,
 }
@@ -90,6 +91,7 @@ impl Command {
             cfg: self.config,
             pre_exec: std::mem::take(&mut self.pre_exec),
             pty_fd: std::mem::take(&mut self.pty_fd),
+            ioctl_denylist: std::mem::take(&mut self.ioctl_denylist),
             seccomp_filter: std::mem::take(&mut self.seccomp_filter),
             seccomp_pipefd,
         });
