@@ -12,7 +12,6 @@ use std::{
     process::{exit, ExitCode},
 };
 
-use libc::{TCGETS, TCSETS, TCSETSF, TCSETSW, TIOCGWINSZ, TIOCSWINSZ};
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
@@ -21,6 +20,7 @@ use nix::{
     sched::{unshare, CloneFlags},
     sys::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
+        prctl::set_dumpable,
         signal::{signal, sigprocmask, SigHandler, SigmaskHow, Signal},
         signalfd::{SfdFlags, SigSet, SignalFd},
         termios::{cfmakeraw, tcgetattr, tcsetattr, OutputFlags, SetArg, Termios},
@@ -29,8 +29,13 @@ use nix::{
 };
 use syd::{
     compat::epoll_ctl_safe,
+    confine::{confine_mdwe, extend_ioctl},
     err::SydResult,
     fs::{closeexcept, duprand, set_exclusive, set_nonblock, winsize_get, winsize_set},
+    ioctl::{
+        TCGETS, TCGETS2, TCSETS, TCSETS2, TCSETSF, TCSETSF2, TCSETSW, TCSETSW2, TIOCGWINSZ,
+        TIOCSWINSZ,
+    },
     landlock_policy::LandlockPolicy,
 };
 
@@ -420,21 +425,41 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
     )?;
 
     // Allow ioctl(2) requests:
-    // 1. TCGETS, aka tcgetattr(3)
-    // 2. TCSETS, aka tcsetattr(3) with TCSANOW
-    // 3. TCSETSW, aka tcsetattr(3) with TCSADRAIN
-    // 4. TCSETSF, aka tcsetattr(3) with TCSAFLUSH
+    // 1. TCGETS{,2}, aka tcgetattr(3)
+    // 2. TCSETS{,2}, aka tcsetattr(3) with TCSANOW
+    // 3. TCSETSW{,2}, aka tcsetattr(3) with TCSADRAIN
+    // 4. TCSETSF{,2}, aka tcsetattr(3) with TCSAFLUSH
     // 5. TIOCGWINSZ, aka winsize_get
     // 6. TIOCSWINSZ, aka winsize_set
+    //
+    // For *2, we use hardcoded versions because not all libcs define them.
+    // They are portable as `struct termios2` has the same size
+    // across 32-bit and 64-bit architectures.
     #[allow(clippy::disallowed_methods)]
     let syscall = ScmpSyscall::from_name("ioctl").unwrap();
     #[allow(clippy::unnecessary_cast)]
-    for request in [TCGETS, TCSETS, TCSETSW, TCSETSF, TIOCGWINSZ, TIOCSWINSZ] {
-        ctx.add_rule_conditional(
-            ScmpAction::Allow,
-            syscall,
-            &[scmp_cmp!($arg1 == request as u64)],
-        )?;
+    for request in [
+        *TCGETS,
+        *TCGETS2,
+        *TCSETS,
+        *TCSETS2,
+        *TCSETSW,
+        *TCSETSW2,
+        *TCSETSF,
+        *TCSETSF2,
+        *TIOCGWINSZ,
+        *TIOCSWINSZ,
+    ] {
+        let request = if let Some(request) = request {
+            request as u64
+        } else {
+            continue;
+        };
+
+        ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == request)])?;
+        if let Some(request) = extend_ioctl(request) {
+            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == request)])?;
+        }
     }
 
     // Allow safe fcntl(2) utility calls.
@@ -486,14 +511,10 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
 
         // Set up Memory-Deny-Write-Execute protections.
         // Ignore errors as PR_SET_MDWE may not be supported.
-        const PR_SET_MDWE: libc::c_int = 65;
-        const PR_MDWE_REFUSE_EXEC_GAIN: libc::c_ulong = 1;
-        // SAFETY: nix has no wrapper for PR_SET_MDWE.
-        let _ = unsafe { libc::prctl(PR_SET_MDWE, PR_MDWE_REFUSE_EXEC_GAIN, 0, 0, 0) };
+        let _ = confine_mdwe(false);
 
         // Set the process dumpable attribute to not-dumpable.
-        // SAFETY: Our nix version does not have a wrapper for set_dumpable yet.
-        let _ = unsafe { libc::prctl(libc::PR_SET_DUMPABLE, 0, 0, 0, 0) };
+        let _ = set_dumpable(false);
 
         // Deny reading the timestamp counter (x86 only).
         // SAFETY: Our nix version does not have a wrapper for SET_TSC yet.
