@@ -10,10 +10,9 @@
 // SPDX-License-Identifier: GPL-3.0
 
 use std::{
-    env,
     mem::MaybeUninit,
     option::Option,
-    os::fd::{AsRawFd, RawFd},
+    os::fd::RawFd,
     sync::{
         atomic::{AtomicBool, Ordering},
         Arc, RwLock,
@@ -21,14 +20,16 @@ use std::{
     thread,
 };
 
-use libseccomp::{scmp_cmp, RawSyscall, ScmpAction, ScmpFilterContext, ScmpSyscall};
-use libseccomp_sys::seccomp_notify_receive;
+#[allow(deprecated)]
+use libc::SOCK_PACKET;
+use libc::{AF_ALG, AF_INET, AF_INET6, AF_NETLINK, AF_UNIX, EACCES, SOCK_RAW};
+use libseccomp::{scmp_cmp, RawSyscall, ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall};
+use libseccomp_sys::{const_scmp_filter_ctx, seccomp_load, seccomp_notify_receive};
 use nix::{
     errno::Errno,
-    poll::PollTimeout,
-    sched::{sched_setaffinity, unshare, CloneFlags, CpuSet},
-    sys::epoll::{Epoll, EpollEvent, EpollFlags},
-    unistd::{close, Pid},
+    fcntl::OFlag,
+    sched::{unshare, CloneFlags},
+    unistd::{close, Gid, Uid},
 };
 
 #[cfg(target_arch = "x86")]
@@ -37,8 +38,11 @@ use crate::cookie::FTRUNCATE64_COOKIE_ARG3;
 use crate::cookie::TRUNCATE64_COOKIE_ARG3;
 use crate::{
     alert,
-    compat::epoll_ctl_safe,
     config::*,
+    confine::{
+        extend_ioctl, scmp_add_setid_rules, scmp_arch_raw, ScmpNotifReq, SydArch, Sydcall,
+        X32_SYSCALL_BIT,
+    },
     cookie::{
         FTRUNCATE64_COOKIE_ARG4, FTRUNCATE64_COOKIE_ARG5, FTRUNCATE_COOKIE_ARG2,
         FTRUNCATE_COOKIE_ARG3, FTRUNCATE_COOKIE_ARG4, FTRUNCATE_COOKIE_ARG5, LINKAT_COOKIE_ARG5,
@@ -53,8 +57,7 @@ use crate::{
         UNLINKAT_COOKIE_ARG5,
     },
     crit,
-    err::{err2no, scmp2no, SydJoinHandle, SydResult},
-    error, extend_ioctl,
+    err::{err2no, SydJoinHandle, SydResult},
     fs::{
         seccomp_notify_respond, AT_EXECVE_CHECK, SECCOMP_IOCTL_NOTIF_ADDFD,
         SECCOMP_IOCTL_NOTIF_LIST, SECCOMP_IOCTL_NOTIF_SEND,
@@ -64,9 +67,8 @@ use crate::{
     path::dotdot_with_nul,
     proc::proc_mmap,
     sandbox::{Flags, Sandbox, SandboxGuard},
-    scmp_arch_raw,
+    timer::AlarmTimer,
     workers::{aes::AesMap, BindMap, WorkerCache, WorkerData},
-    ExportMode, ScmpNotifReq, SydArch, Sydcall,
 };
 
 const EOWNERDEAD: i32 = -libc::EOWNERDEAD;
@@ -75,7 +77,6 @@ const EOWNERDEAD: i32 = -libc::EOWNERDEAD;
 pub(crate) struct Worker {
     fd: RawFd,
     flags: Flags,
-    epoll: Arc<Epoll>,
     cache: Arc<WorkerCache<'static>>,
     sandbox: Arc<RwLock<Sandbox>>,
     handlers: Arc<HandlerMap>,
@@ -90,7 +91,6 @@ impl Worker {
     #[allow(clippy::too_many_arguments)]
     pub(crate) fn new(
         fd: RawFd,
-        epoll: Arc<Epoll>,
         cache: Arc<WorkerCache<'static>>,
         sandbox: Arc<RwLock<Sandbox>>,
         handlers: Arc<HandlerMap>,
@@ -107,7 +107,6 @@ impl Worker {
         Worker {
             fd,
             flags,
-            epoll,
             cache,
             sandbox,
             handlers,
@@ -120,7 +119,18 @@ impl Worker {
     }
 
     #[allow(clippy::cognitive_complexity)]
-    pub(crate) fn try_spawn(self, mut confined: bool) -> Result<SydJoinHandle<()>, Errno> {
+    pub(crate) fn try_spawn(
+        self,
+        ctx: Option<&ScmpFilterContext>,
+    ) -> Result<SydJoinHandle<()>, Errno> {
+        // SAFETY: ScmpFilterContext is not Send,
+        // so we cannot pass it between threads.
+        // Therefore we pass a pointer which is
+        // owned by the monitor thread. This
+        // pointer is guaranteed to be valid
+        // throughout Syd's lifetime.
+        let mut ctx = ctx.map(|ctx| ctx.as_ptr() as usize);
+
         thread::Builder::new()
             .name("syd_emu".to_string())
             .stack_size(EMU_STACK_SIZE)
@@ -166,91 +176,60 @@ impl Worker {
                     std::process::exit(101);
                 }
 
+                // SAFETY: We use exit_group(2) here to bail,
+                // because expiring idle threads is a critical safety feature.
+                let mut timer = if let Some(keep_alive) = self.keep_alive {
+                    match AlarmTimer::new(keep_alive) {
+                        Ok(timer) => Some(timer),
+                        Err(errno) => {
+                            alert!("ctx": "boot", "op": "timer_create_for_emulator_thread",
+                                "msg": format!("failed to set up timer: {errno}"),
+                                "err": errno as i32);
+                            std::process::exit(101);
+                        }
+                    }
+                } else {
+                    None
+                };
+
                 // Create sentinel, that will handle graceful teardown.
                 let mut sentinel = Sentinel::new(&self);
 
                 // Thread successfully started, increment total worker count.
-                let worker_count = self.worker_data.increment_worker_total();
+                self.worker_data.increment_worker_total();
 
-                // Attempt to set thread's CPU affinity mask.
-                // We pin the main, init and monitor threads to CPU:0.
-                // Emulator threads are pinned based on num-cpus.
-                #[allow(clippy::arithmetic_side_effects)]
-                let cpu_id = worker_count % *NPROC;
-                let mut cpu_set = CpuSet::new();
-                if cpu_set.set(cpu_id).is_ok() {
-                    match sched_setaffinity(Pid::from_raw(0), &cpu_set) {
-                        Ok(_) => {
-                            info!("ctx": "boot", "op": "pin_emulator_thread",
-                                "msg": format!("pinned emulator thread to CPU:{cpu_id}"),
-                                "cpu": cpu_id);
-                        }
-                        Err(errno) => {
-                            error!("ctx": "boot", "op": "pin_emulator_thread",
-                                "msg": format!("failed to pin emulator thread to CPU:{cpu_id}: {errno}"),
-                                "err": errno as i32,
-                                "cpu": cpu_id);
-                        }
-                    }
-                }
-
-                let dry_run = env::var_os(ENV_SKIP_SCMP).is_some() || ExportMode::from_env().is_some();
                 loop {
-                    // SAFETY: Confine if/once locked.
-                    if !confined && Sandbox::locked_once() {
-                        // SAFETY: We use exit_group(2) here to bail,
-                        // because this confinement is a critical safety feature.
-                        let ctx = match Self::prepare_confine(
-                            self.fd,
-                            self.epoll.0.as_raw_fd(),
-                            self.flags,
-                            is_crypt,
-                        ) {
-                            Ok(ctx) => ctx,
-                            Err(error) => {
-                                let errno = error.errno().unwrap_or(Errno::ENOSYS);
+                    // Confine and drop filter if sandbox is locked.
+                    if let Some(filter) = ctx {
+                        if Sandbox::locked_once() {
+                            // SAFETY: filter pointer is owned by the
+                            // monitor thread and is valid for Syd's
+                            // lifetime.
+                            let error = unsafe { seccomp_load(filter as const_scmp_filter_ctx) };
+
+                            // SAFETY: We use exit_group(2) here to bail,
+                            // because this confinement is a critical safety feature.
+                            if error != 0 {
+                                let errno = Errno::from_raw(error.abs());
                                 alert!("ctx": "boot", "op": "confine_emulator_thread",
                                     "msg": format!("failed to confine: {error}"),
                                     "err": errno as i32);
                                 std::process::exit(101);
                             }
-                        };
 
-                        // Honour dry-run when exporting.
-                        if !dry_run {
-                            let safe_setid = self.flags.intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
+                            let safe_setid = self.flags.intersects(
+                                Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID,
+                            );
                             info!("ctx": "confine", "op": "confine_emulator_thread",
                                 "msg": format!("emulator thread confined with{} SROP mitigation",
                                     if safe_setid { "out" } else { "" }));
 
-                            // SAFETY: We use exit_group(2) here to bail,
-                            // because this confinement is a critical safety feature.
-                            if let Err(error) = ctx.load() {
-                                let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
-                                alert!("ctx": "boot", "op": "confine_emulator_thread",
-                                    "msg": format!("failed to confine: {error}"),
-                                    "err": errno as i32);
-                                std::process::exit(101);
-                            }
-                        } else {
-                            error!("ctx": "confine", "op": "confine_emulator_thread",
-                                "msg": "emulator thread is running unconfined in debug mode");
+                            ctx = None;
                         }
-
-                        confined = true;
-                    }
-
-                    // Wait for the request to become ready as necessary.
-                    // epoll_wait(2) will timeout and exit for non-core threads.
-                    if self.poll().is_err() {
-                        // Timeout or critical error.
-                        // Decrement worker total and exit.
-                        self.worker_data.decrement_worker_total();
-                        break;
                     }
 
                     // Receive seccomp notification.
-                    let request = if let Ok(request) = self.receive() {
+                    let request = if let Ok(request) = self.receive(&mut timer) {
                         request
                     } else {
                         // Critical error, decrement worker total and exit.
@@ -273,30 +252,47 @@ impl Worker {
                 }
 
                 Ok(())
-            }).map_err(|err| err2no(&err))
+            })
+            .map_err(|err| err2no(&err))
     }
 
-    fn receive(&self) -> Result<Option<ScmpNotifReq>, Errno> {
+    fn receive(&self, timer: &mut Option<AlarmTimer>) -> Result<Option<ScmpNotifReq>, Errno> {
         // Receive and return request.
         // Break if file descriptor was closed.
         // Ignore rest of the errors as we cannot handle them,
         // e.g: EINTR|ENOENT: task is killed mid-way.
-        match self.read() {
+        match self.read(timer) {
             Ok(request) => Ok(Some(request)),
             Err(Errno::EBADF) => Err(Errno::EBADF),
+            Err(Errno::EINTR) if timer.is_some() => Err(Errno::EINTR),
             Err(_) => Ok(None),
         }
     }
 
-    fn read(&self) -> Result<ScmpNotifReq, Errno> {
+    fn read(&self, timer: &mut Option<AlarmTimer>) -> Result<ScmpNotifReq, Errno> {
         // Use libc::seccomp_notif rather than libseccomp_sys's.
         // The latter is opaque and requires us to do a heap
         // allocation which we don't always want.
         let mut req: MaybeUninit<libc::seccomp_notif> = MaybeUninit::zeroed();
 
+        if let Some(timer) = timer {
+            // SAFETY: Something is awfully wrong if we cannot
+            // set the timer so we panic here to indicate this.
+            #[allow(clippy::disallowed_methods)]
+            timer.start().expect("timer_settime");
+        }
+
         // SAFETY: libseccomp's wrapper allocates each call.
-        // Note: EINTR means child killed by signal!
-        Errno::result(unsafe { seccomp_notify_receive(self.fd, req.as_mut_ptr().cast()) })?;
+        // Note: EINTR may also mean child killed by signal!
+        let res =
+            Errno::result(unsafe { seccomp_notify_receive(self.fd, req.as_mut_ptr().cast()) });
+
+        if let Some(timer) = timer {
+            // SAFETY: See above.
+            #[allow(clippy::disallowed_methods)]
+            timer.stop().expect("timer_settime");
+        }
+        res?;
 
         // SAFETY: seccomp_notify_receive returned success.
         // Request is populated and accessing it is safe.
@@ -311,7 +307,14 @@ impl Worker {
     }
 
     #[allow(clippy::cognitive_complexity)]
-    fn handle(&self, req: ScmpNotifReq) {
+    fn handle(&self, mut req: ScmpNotifReq) {
+        // Correct architecture for x32 syscalls.
+        if req.data.arch == ScmpArch::X8664
+            && req.data.syscall.as_raw_syscall() & X32_SYSCALL_BIT != 0
+        {
+            req.data.arch = ScmpArch::X32;
+        }
+
         // Lookup the system call handler, panic if not found.
         let syscall = Sydcall(req.data.syscall, scmp_arch_raw(req.data.arch));
         let handler = if let Some(handler) = self.handlers.get(&syscall) {
@@ -360,12 +363,6 @@ impl Worker {
 
         // See above.
         if response.error == EOWNERDEAD {
-            // Note, threads blocked on epoll_wait will not
-            // wake up even if we close the epoll fd or
-            // delete the seccomp fd from epoll wait-list here.
-            // That said, they'll never ever wake up again,
-            // and therefore will not consume system resources.
-            let _ = epoll_ctl_safe(&self.epoll.0, self.fd, None);
             let _ = close(self.fd);
 
             // Inform the monitor thread to exit.
@@ -373,61 +370,14 @@ impl Worker {
         }
     }
 
-    fn poll(&self) -> Result<(), Errno> {
-        let timeout = if let Some(keep_alive) = self.keep_alive {
-            PollTimeout::from(keep_alive)
-        } else {
-            PollTimeout::NONE
-        };
-
-        // Wait for an event and handle EINTR.
-        // Retire threads which have exited along the way.
-        let mut events = [EpollEvent::empty(); 1];
-        loop {
-            if self.should_exit.load(Ordering::Relaxed) {
-                // Exit notified, do not try to wait on epoll again.
-                return Err(Errno::ESRCH);
-            }
-
-            match self.epoll.wait(&mut events, timeout) {
-                Ok(0) if self.keep_alive.is_some() => return Err(Errno::ETIMEDOUT),
-                Ok(0) | Err(Errno::EINTR) => {} // try again.
-                Ok(1) if events[0].data() == 0 => {
-                    // Seccomp request.
-                    let flags = events[0].events();
-
-                    return if flags.contains(EpollFlags::EPOLLIN) {
-                        // Received scmp request.
-                        Ok(())
-                    } else if flags.intersects(
-                        EpollFlags::EPOLLERR | EpollFlags::EPOLLHUP | EpollFlags::EPOLLRDHUP,
-                    ) {
-                        // Exit on any hang-up or error.
-                        //
-                        // Inform the monitor thread to exit.
-                        self.should_exit.store(true, Ordering::Relaxed);
-
-                        Err(Errno::EPIPE)
-                    } else {
-                        unreachable!(
-                            "BUG: seccomp request event with invalid flags {:#x}!",
-                            flags.bits()
-                        );
-                    };
-                }
-                Ok(n) => unreachable!("BUG: epoll_wait returned ${n} unexpectedly!"),
-                Err(errno) => return Err(errno),
-            };
-        }
-    }
-
     /// Confine Worker thread.
     #[allow(clippy::cognitive_complexity)]
     pub(crate) fn prepare_confine(
         seccomp_fd: RawFd,
-        epoll_fd: RawFd,
         flags: Flags,
         is_crypt: bool,
+        transit_uids: &[(Uid, Uid)],
+        transit_gids: &[(Gid, Gid)],
     ) -> SydResult<ScmpFilterContext> {
         // Create seccomp filter with default action.
         let mut ctx = ScmpFilterContext::new(ScmpAction::KillProcess)?;
@@ -438,7 +388,7 @@ impl Worker {
 
         // Disable Speculative Store Bypass mitigations
         // with trace/allow_unsafe_spec_exec:1
-        ctx.set_ctl_ssb(flags.contains(Flags::FL_ALLOW_UNSAFE_SPEC_EXEC))?;
+        ctx.set_ctl_ssb(flags.allow_unsafe_spec_exec())?;
 
         // DO NOT synchronize filter to all threads.
         // Main thread will confine itself.
@@ -476,7 +426,6 @@ impl Worker {
         // Syscall argument cookies for SECCOMP_IOCTL_NOTIF_SEND my be disabled
         // at startup with trace/allow_unsafe_nocookie:1.
         let sysname = "ioctl";
-        let restrict_cookie = !flags.contains(Flags::FL_ALLOW_UNSAFE_NOCOOKIE);
         #[allow(clippy::cast_sign_loss)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
@@ -489,7 +438,8 @@ impl Worker {
                     // and aims steal file descriptors.
                     //
                     // Randomizing the seccomp-fd at startup is another mitigation against this.
-                    if restrict_cookie && *ioctl_request == SECCOMP_IOCTL_NOTIF_ADDFD {
+                    if !flags.allow_unsafe_nocookie() && *ioctl_request == SECCOMP_IOCTL_NOTIF_ADDFD
+                    {
                         rules.extend(&[
                             scmp_cmp!($arg3 == (*SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG3).into()),
                             scmp_cmp!($arg4 == (*SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG4).into()),
@@ -503,7 +453,8 @@ impl Worker {
                     // response in order to pass-through a system call to the host Linux kernel.
                     //
                     // Randomizing the seccomp-fd at startup is another mitigation against this.
-                    if restrict_cookie && *ioctl_request == SECCOMP_IOCTL_NOTIF_SEND {
+                    if !flags.allow_unsafe_nocookie() && *ioctl_request == SECCOMP_IOCTL_NOTIF_SEND
+                    {
                         rules.extend(&[
                             scmp_cmp!($arg3 == (*SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG3).into()),
                             scmp_cmp!($arg4 == (*SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG4).into()),
@@ -527,24 +478,6 @@ impl Worker {
             }
         }
 
-        // Allow epoll(7) API to our single epoll fd only.
-        #[allow(clippy::cast_sign_loss)]
-        for sysname in EPOLL_SYSCALLS {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule_conditional(
-                        ScmpAction::Allow,
-                        syscall,
-                        &[scmp_cmp!($arg0 == epoll_fd as u64)],
-                    )?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_emu_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
         // Allow unshare(2) with CLONE_FS|CLONE_FILES only.
         let sysname = "unshare";
         match ScmpSyscall::from_name(sysname) {
@@ -562,37 +495,22 @@ impl Worker {
             }
         }
 
-        // SAFETY: Allow write(2) _only_ if trace/memory_access allows
-        // proc_pid_mem(5) access in which case we cannot avoid it.
-        // The default is to use cross memory attach with
-        // process_vm_{read,write}v(2) unconditionally since
-        // version 3.32.6 unless relaxed.
-        let log_fd = crate::log::LOG_FD.load(std::sync::atomic::Ordering::Relaxed);
-        for sysname in ["write", "writev", "fdatasync"] {
-            #[allow(clippy::cast_sign_loss)]
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) if cfg!(feature = "prof") || Sandbox::memory_access() < 2 => {
-                    // Allow write(2) globally for proc_pid_mem(5) access.
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Ok(syscall) => {
-                    // Allow write(2) access to syd::log::LOG_FD only.
-                    // If logging is disabled, deny write(2) completely.
-                    // Note, we do allow the use-case where LOG_FD is
-                    // intentionally set to a negative value to disable
-                    // logging, hence we cast directly to an u64 and not
-                    // use u64::try_from!
-                    ctx.add_rule_conditional(
-                        ScmpAction::Allow,
-                        syscall,
-                        &[scmp_cmp!($arg0 == log_fd as u64)],
-                    )?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_emu_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
+        // Deny pipe2(2) O_NOTIFICATION_PIPE flag.
+        let sysname = "pipe2";
+        #[allow(clippy::cast_sign_loss)]
+        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
+            // O_NOTIFICATION_PIPE is equivalent to O_EXCL,
+            // see: linux/watch_queue.h
+            const O_NOTIFICATION_PIPE: u64 = OFlag::O_EXCL.bits() as u64;
+
+            ctx.add_rule_conditional(
+                ScmpAction::Allow,
+                syscall,
+                &[scmp_cmp!($arg1 & O_NOTIFICATION_PIPE == 0)],
+            )?;
+        } else {
+            info!("ctx": "confine", "op": "allow_emu_syscall",
+                "msg": format!("invalid or unsupported syscall {sysname}"));
         }
 
         // Allow safe fcntl(2) utility calls.
@@ -642,7 +560,7 @@ impl Worker {
 
         // Syscall argument cookies may be disabled
         // at startup with trace/allow_unsafe_nocookie:1.
-        if restrict_cookie {
+        if !flags.allow_unsafe_nocookie() {
             // memfd_create(2) may be used only with syscall argument cookies.
             let sysname = "memfd_create";
             #[allow(clippy::useless_conversion)]
@@ -838,37 +756,6 @@ impl Worker {
                         "msg": format!("invalid or unsupported syscall {sysname}"));
                 }
             }
-
-            // socket(2) may be used only with syscall argument cookies.
-            // Note: We only enforce this on architectures where the
-            // system call is direct and there's no socketcall(2)
-            // multiplexer indirection.
-            #[allow(clippy::cast_possible_truncation)]
-            #[allow(clippy::useless_conversion)]
-            if let Some(syscall) =
-                SYS_SOCKET.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall))
-            {
-                ctx.add_rule_conditional(
-                    ScmpAction::Allow,
-                    syscall,
-                    &[
-                        scmp_cmp!($arg3 == (*SOCKET_COOKIE_ARG3).into()),
-                        scmp_cmp!($arg4 == (*SOCKET_COOKIE_ARG4).into()),
-                        scmp_cmp!($arg5 == (*SOCKET_COOKIE_ARG5).into()),
-                    ],
-                )?;
-            } else {
-                match ScmpSyscall::from_name("socket") {
-                    Ok(syscall) => {
-                        // Allow socketcall(2).
-                        ctx.add_rule(ScmpAction::Allow, syscall)?;
-                    }
-                    Err(_) => {
-                        info!("ctx": "confine", "op": "allow_emu_syscall",
-                            "msg": "invalid or unsupported syscall socket");
-                    }
-                }
-            }
         } else {
             // trace/allow_unsafe_nocookie: Allow access without cookies.
             for sysname in [
@@ -881,7 +768,6 @@ impl Worker {
                 "linkat",
                 "unlinkat",
                 "openat2",
-                "socket",
             ] {
                 match ScmpSyscall::from_name(sysname) {
                     Ok(syscall) => {
@@ -934,6 +820,71 @@ impl Worker {
             }
         }
 
+        // socket(2) may be used only with syscall argument cookies.
+        // Note: We only enforce this on architectures where the
+        // system call is direct and there's no socketcall(2)
+        // multiplexer indirection.
+        #[allow(clippy::cast_possible_truncation)]
+        #[allow(clippy::cast_sign_loss)]
+        #[allow(clippy::useless_conversion)]
+        #[allow(deprecated)]
+        if let Some(syscall) = SYS_SOCKET.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
+            // Deny based on socket type.
+            if !flags.allow_unsafe_socket() {
+                for ty in [SOCK_RAW as u64, SOCK_PACKET as u64] {
+                    ctx.add_rule_conditional(
+                        ScmpAction::Errno(EACCES),
+                        syscall,
+                        &[scmp_cmp!($arg1 == ty)],
+                    )?;
+                }
+            }
+
+            // Allow only specified socket domains.
+            let domains = if !flags.allow_unsupp_socket() {
+                let mut domains = vec![AF_UNIX, AF_INET, AF_INET6, AF_NETLINK];
+                if flags.allow_safe_kcapi() {
+                    domains.push(AF_ALG);
+                }
+                Some(domains)
+            } else {
+                None
+            };
+
+            // Secure using syscall argument cookies.
+            let mut rules = vec![];
+            if !flags.allow_unsafe_nocookie() {
+                rules.extend(&[
+                    scmp_cmp!($arg3 == (*SOCKET_COOKIE_ARG3).into()),
+                    scmp_cmp!($arg4 == (*SOCKET_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*SOCKET_COOKIE_ARG5).into()),
+                ]);
+            }
+
+            if let Some(domains) = domains {
+                for domain in domains {
+                    rules.push(scmp_cmp!($arg0 == domain as u64));
+                    ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
+                    rules.pop();
+                }
+            } else if rules.is_empty() {
+                ctx.add_rule(ScmpAction::Allow, syscall)?;
+            } else {
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
+            }
+        } else {
+            match ScmpSyscall::from_name("socket") {
+                Ok(syscall) => {
+                    // Allow socketcall(2).
+                    ctx.add_rule(ScmpAction::Allow, syscall)?;
+                }
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_emu_syscall",
+                        "msg": "invalid or unsupported syscall socket");
+                }
+            }
+        }
+
         // Allow futex system calls.
         for sysname in FUTEX_SYSCALLS {
             match ScmpSyscall::from_name(sysname) {
@@ -980,22 +931,18 @@ impl Worker {
             }
         }
 
-        if flags.intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID) {
-            // SAFETY: Main thread confines these further.
-            // As these system calls as per-process,
-            // the main thread's seccomp rules will apply
-            // to us even without TSYNC.
-            for sysname in SET_ID_SYSCALLS {
-                match ScmpSyscall::from_name(sysname) {
-                    Ok(syscall) => {
-                        ctx.add_rule(ScmpAction::Allow, syscall)?;
-                    }
-                    Err(_) => {
-                        info!("ctx": "confine", "op": "allow_emu_syscall",
-                            "msg": format!("invalid or unsupported syscall {sysname}"));
-                    }
-                }
-            }
+        // Allow UID/GID changing system calls as necessary.
+        let safe_setuid = flags.allow_safe_setuid();
+        let safe_setgid = flags.allow_safe_setgid();
+        if safe_setuid || safe_setgid {
+            scmp_add_setid_rules(
+                "emu",
+                &mut ctx,
+                safe_setuid,
+                safe_setgid,
+                transit_uids,
+                transit_gids,
+            )?;
 
             // SAFETY:
             // 1. cap{g,s}et is necessary to drop CAP_SET{U,G}ID after changing {U,G}ID.
