@@ -20,7 +20,7 @@ use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
     sched::{unshare, CloneFlags},
-    unistd::{getpid, write, Pid},
+    unistd::{getpid, write, Gid, Pid, Uid},
 };
 use serde::{ser::SerializeMap, Serialize, Serializer};
 
@@ -28,15 +28,17 @@ use crate::{
     alert,
     cache::SysInterrupt,
     config::*,
+    confine::{scmp_add_setid_rules, ExportMode},
     cookie::{OPENAT2_COOKIE_ARG4, OPENAT2_COOKIE_ARG5},
     crit,
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
+    error,
     fs::{closeexcept, retry_on_eintr, seccomp_notify_id_valid, tgkill},
     info,
     proc::proc_status,
     sandbox::Flags,
+    sigset::SydSigSet,
     workers::WorkerCache,
-    ExportMode, SydSigSet,
 };
 
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
@@ -68,8 +70,12 @@ impl Serialize for InterruptSource {
 
 #[derive(Clone)]
 pub(crate) struct Interrupter {
-    flags: Flags,
     seccomp_fd: RawFd,
+    flags: Flags,
+
+    transit_uids: Vec<(Uid, Uid)>,
+    transit_gids: Vec<(Gid, Gid)>,
+
     should_exit: Arc<AtomicBool>,
     cache: Arc<WorkerCache<'static>>,
 }
@@ -78,6 +84,8 @@ impl Interrupter {
     pub(crate) fn new(
         seccomp_fd: RawFd,
         flags: Flags,
+        transit_uids: &[(Uid, Uid)],
+        transit_gids: &[(Gid, Gid)],
         should_exit: Arc<AtomicBool>,
         cache: Arc<WorkerCache<'static>>,
     ) -> Self {
@@ -86,6 +94,8 @@ impl Interrupter {
             seccomp_fd,
             should_exit,
             cache,
+            transit_uids: transit_uids.to_vec(),
+            transit_gids: transit_gids.to_vec(),
         }
     }
 
@@ -158,6 +168,7 @@ impl Interrupter {
                         "err": errno as i32);
                     std::process::exit(101);
                 }
+                drop(set);
 
                 // To be used by tgkill when signaling threads.
                 let tgid = getpid();
@@ -165,37 +176,50 @@ impl Interrupter {
                 // Honour dry-run when exporting.
                 let dry_run =
                     env::var_os(ENV_SKIP_SCMP).is_some() || ExportMode::from_env().is_some();
-                let safe_setid = self
-                    .flags
-                    .intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
 
                 // Confine `syd_int' thread.
-                // SAFETY: We use exit_group(2) here to bail,
-                // because this confinement is a critical safety feature.
-                let ctx = match Self::prepare_confine(self.seccomp_fd, tgid, self.flags, dry_run) {
-                    Ok(ctx) => ctx,
-                    Err(error) => {
-                        let errno = error.errno().unwrap_or(Errno::ENOSYS);
+                if !dry_run {
+                    // SAFETY: We use exit_group(2) here to bail,
+                    // because this confinement is a critical safety feature.
+                    let ctx = match Self::prepare_confine(
+                        self.seccomp_fd,
+                        tgid,
+                        self.flags,
+                        &self.transit_uids,
+                        &self.transit_gids,
+                        false,
+                    ) {
+                        Ok(ctx) => ctx,
+                        Err(error) => {
+                            let errno = error.errno().unwrap_or(Errno::ENOSYS);
+                            alert!("ctx": "boot", "op": "confine_interrupt_thread",
+                                "msg": format!("failed to confine: {error}"),
+                                "err": errno as i32);
+                            std::process::exit(101);
+                        }
+                    };
+
+                    // Load seccomp(2) BPF into the kernel.
+                    // SAFETY: We use exit_group(2) here to bail,
+                    // because this confinement is a critical safety feature.
+                    if let Err(error) = ctx.load() {
+                        let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
                         alert!("ctx": "boot", "op": "confine_interrupt_thread",
                             "msg": format!("failed to confine: {error}"),
                             "err": errno as i32);
                         std::process::exit(101);
                     }
-                };
 
-                // Load seccomp(2) BPF into the kernel.
-                // SAFETY: We use exit_group(2) here to bail,
-                // because this confinement is a critical safety feature.
-                if let Err(error) = ctx.load() {
-                    let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
-                    alert!("ctx": "boot", "op": "confine_interrupt_thread",
-                        "msg": format!("failed to confine: {error}"),
-                        "err": errno as i32);
-                    std::process::exit(101);
+                    let safe_setid = self
+                        .flags
+                        .intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
+                    info!("ctx": "confine", "op": "confine_interrupt_thread",
+                        "msg": format!("interrupt thread confined with{} SROP mitigation",
+                            if safe_setid { "out" } else { "" }));
+                } else {
+                    error!("ctx": "confine", "op": "confine_int_thread",
+                        "msg": "interrupt thread is running unconfined in debug mode");
                 }
-                info!("ctx": "confine", "op": "confine_interrupt_thread",
-                    "msg": format!("interrupt thread confined with{} SROP mitigation",
-                        if safe_setid { "out" } else { "" }));
 
                 // Enter main loop.
                 self.main(tgid)
@@ -343,6 +367,8 @@ impl Interrupter {
         seccomp_fd: RawFd,
         tgid: Pid,
         flags: Flags,
+        transit_uids: &[(Uid, Uid)],
+        transit_gids: &[(Gid, Gid)],
         _dry_run: bool,
     ) -> SydResult<ScmpFilterContext> {
         // Note, we cannot confine `syd_int` with a per-thread
@@ -359,7 +385,7 @@ impl Interrupter {
 
         // Disable Speculative Store Bypass mitigations
         // with trace/allow_unsafe_spec_exec:1
-        ctx.set_ctl_ssb(flags.contains(Flags::FL_ALLOW_UNSAFE_SPEC_EXEC))?;
+        ctx.set_ctl_ssb(flags.allow_unsafe_spec_exec())?;
 
         // DO NOT synchronize filter to all threads.
         // Other threads will self-confine.
@@ -500,7 +526,7 @@ impl Interrupter {
         // use u64::try_from!
         // Exception: Allow write(2) globally if profiling is enabled.
         let log_fd = crate::log::LOG_FD.load(Ordering::Relaxed);
-        for sysname in ["write", "writev", "fdatasync"] {
+        for sysname in ["write", "writev"] {
             #[allow(clippy::cast_sign_loss)]
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) if cfg!(feature = "prof") => {
@@ -559,36 +585,18 @@ impl Interrupter {
             }
         }
 
-        if flags.intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID) {
-            // SAFETY: Main thread confines these further.
-            // As these system calls as per-process,
-            // the main thread's seccomp rules will apply
-            // to us even without TSYNC.
-            for sysname in SET_ID_SYSCALLS {
-                match ScmpSyscall::from_name(sysname) {
-                    Ok(syscall) => {
-                        ctx.add_rule(ScmpAction::Allow, syscall)?;
-                    }
-                    Err(_) => {
-                        info!("ctx": "confine", "op": "allow_int_syscall",
-                            "msg": format!("invalid or unsupported syscall {sysname}"));
-                    }
-                }
-            }
-
-            // SAFETY:
-            // Signal system calls are necessary to handle reserved signals.
-            for sysname in ["sigreturn", "rt_sigreturn"] {
-                match ScmpSyscall::from_name(sysname) {
-                    Ok(syscall) => {
-                        ctx.add_rule(ScmpAction::Allow, syscall)?;
-                    }
-                    Err(_) => {
-                        info!("ctx": "confine", "op": "allow_int_syscall",
-                            "msg": format!("invalid or unsupported syscall {sysname}"));
-                    }
-                }
-            }
+        // Allow UID/GID changing system calls as necessary.
+        let safe_setuid = flags.allow_safe_setuid();
+        let safe_setgid = flags.allow_safe_setgid();
+        if safe_setuid || safe_setgid {
+            scmp_add_setid_rules(
+                "int",
+                &mut ctx,
+                safe_setuid,
+                safe_setgid,
+                transit_uids,
+                transit_gids,
+            )?;
         }
 
         Ok(ctx)
