@@ -8,7 +8,6 @@
 
 use std::{
     borrow::Cow,
-    collections::HashMap,
     hash::{Hash, Hasher},
     io::{IoSlice, IoSliceMut},
     os::{
@@ -44,10 +43,11 @@ use crate::{
     alert,
     compat::{epoll_ctl_mod_safe, epoll_ctl_safe},
     config::*,
+    confine::scmp_add_setid_rules,
     crit,
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
     fs::{closeexcept, duprand, peer_cred, retry_on_eintr},
-    hash::SydRandomState,
+    hash::SydHashMap,
     info,
     landlock_policy::LandlockPolicy,
     path::{XPath, XPathBuf},
@@ -77,6 +77,10 @@ pub(crate) struct IpcWorker {
     flags: Flags,
     // Credentials for authentication.
     creds: (Option<Uid>, Option<Gid>),
+
+    // SafeSetId UID/GID transitions.
+    transit_uids: Vec<(Uid, Uid)>,
+    transit_gids: Vec<(Gid, Gid)>,
 }
 
 // Prompt modes supported (similar to HAProxy)
@@ -125,7 +129,7 @@ struct Connection {
 
 // Type for Connection map.
 // We use HashMap because HashSet does not have get_mut!
-type ConnectionMap = HashMap<RawFd, Connection, SydRandomState>;
+type ConnectionMap = SydHashMap<RawFd, Connection>;
 
 impl Hash for Connection {
     fn hash<H: Hasher>(&self, state: &mut H) {
@@ -223,7 +227,14 @@ impl IpcWorker {
     //
     // This does not bind or listen yet; only stores the path.
     // Returns immediately without error.
-    pub(crate) fn new(addr: &XPath, uid: Option<Uid>, gid: Option<Gid>, flags: Flags) -> Self {
+    pub(crate) fn new(
+        addr: &XPath,
+        uid: Option<Uid>,
+        gid: Option<Gid>,
+        flags: Flags,
+        transit_uids: &[(Uid, Uid)],
+        transit_gids: &[(Gid, Gid)],
+    ) -> Self {
         IpcWorker {
             flags,
             creds: (uid, gid),
@@ -231,6 +242,8 @@ impl IpcWorker {
             epoll: None,
             sock: None,
             sandbox: None,
+            transit_uids: transit_uids.to_vec(),
+            transit_gids: transit_gids.to_vec(),
         }
     }
 
@@ -319,6 +332,8 @@ impl IpcWorker {
     pub(crate) fn prepare_confine(
         epoll_fd: RawFd,
         flags: Flags,
+        transit_uids: &[(Uid, Uid)],
+        transit_gids: &[(Gid, Gid)],
         dry_run: bool,
     ) -> SydResult<ScmpFilterContext> {
         if !dry_run {
@@ -347,7 +362,7 @@ impl IpcWorker {
 
         // Disable Speculative Store Bypass mitigations
         // with trace/allow_unsafe_spec_exec:1
-        ctx.set_ctl_ssb(flags.contains(Flags::FL_ALLOW_UNSAFE_SPEC_EXEC))?;
+        ctx.set_ctl_ssb(flags.allow_unsafe_spec_exec())?;
 
         // We kill for bad system call and bad arch.
         ctx.set_act_badarch(ScmpAction::KillProcess)?;
@@ -464,7 +479,7 @@ impl IpcWorker {
         // use u64::try_from!
         // Exception: Allow write(2) globally if profiling is enabled.
         let log_fd = crate::log::LOG_FD.load(Ordering::Relaxed);
-        for sysname in ["write", "writev", "fdatasync"] {
+        for sysname in ["write", "writev"] {
             #[allow(clippy::cast_sign_loss)]
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) if cfg!(feature = "prof") => {
@@ -523,36 +538,18 @@ impl IpcWorker {
             }
         }
 
-        if flags.intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID) {
-            // SAFETY: Main thread confines these further.
-            // As these system calls as per-process,
-            // the main thread's seccomp rules will apply
-            // to us even without TSYNC.
-            for sysname in SET_ID_SYSCALLS {
-                match ScmpSyscall::from_name(sysname) {
-                    Ok(syscall) => {
-                        ctx.add_rule(ScmpAction::Allow, syscall)?;
-                    }
-                    Err(_) => {
-                        info!("ctx": "confine", "op": "allow_ipc_syscall",
-                            "msg": format!("invalid or unsupported syscall {sysname}"));
-                    }
-                }
-            }
-
-            // SAFETY:
-            // Signal system calls are necessary to handle reserved signals.
-            for sysname in ["sigreturn", "rt_sigreturn"] {
-                match ScmpSyscall::from_name(sysname) {
-                    Ok(syscall) => {
-                        ctx.add_rule(ScmpAction::Allow, syscall)?;
-                    }
-                    Err(_) => {
-                        info!("ctx": "confine", "op": "allow_ipc_syscall",
-                            "msg": format!("invalid or unsupported syscall {sysname}"));
-                    }
-                }
-            }
+        // Allow UID/GID changing system calls as necessary.
+        let safe_setuid = flags.allow_safe_setuid();
+        let safe_setgid = flags.allow_safe_setgid();
+        if safe_setuid || safe_setgid {
+            scmp_add_setid_rules(
+                "ipc",
+                &mut ctx,
+                safe_setuid,
+                safe_setgid,
+                transit_uids,
+                transit_gids,
+            )?;
         }
 
         Ok(ctx)
@@ -646,7 +643,13 @@ impl IpcWorker {
         // Confine `syd_ipc' thread.
         // SAFETY: We use exit_group(2) here to bail,
         // because this confinement is a critical safety feature.
-        let ctx = match Self::prepare_confine(epoll.0.as_raw_fd(), self.flags, false) {
+        let ctx = match Self::prepare_confine(
+            epoll.0.as_raw_fd(),
+            self.flags,
+            &self.transit_uids,
+            &self.transit_gids,
+            false,
+        ) {
             Ok(ctx) => ctx,
             Err(error) => {
                 let errno = error.errno().unwrap_or(Errno::ENOSYS);
@@ -667,6 +670,8 @@ impl IpcWorker {
                 "err": errno as i32);
             std::process::exit(101);
         }
+        drop(ctx);
+
         info!("ctx": "confine", "op": "confine_ipc_thread",
             "msg": "IPC thread confined");
 
